/*
 * eip-constants.h
 * 
 * Constants used in EIP processing
 * 
 */



#ifndef __EIP_CONSTANTS_H__
#define __EIP_CONSTANTS_H__

#include <stdint.h>

// constant macros
#define EIP_TYPE_SETUP 1  // type 1 datagram sets up a new flow
#define EIP_TYPE_REUSE 2  // type 2 datagram reuses an existing flow
#define EIP_TYPE_RENEW 3  // type 3 datagram renews an existing flow
#define BYTE 0xFF         // mask for a single byte
#define IPHDR_ADDR_OFFSET 12 // bytes from start of IP header to src/dst address pair
#define EIP_VALID 1       // flag = true, indicating validity check has passed
#define EIP_INVALID 0     // flag = false, indicating validity check has failed
#define EIP_CONT_YES 1    // flag = true, incating continuation is present / should be added
#define EIP_CONT_NO 0     // flag = false, incating continuation is not present / does not need to be added
// protocol parameter macros
#define EIP_TADDRESS_VTIME 60 // t-address time validity constraint (seconds from generation time)
#define EIP_TADDRESS_VBYTES 1024000 // t-address bytes validity constraint (number of bytes allowed)
//#define EIP_TADDRESS_VBYTES 16777216 // t-address bytes validity constraint (number of bytes allowed)
#define EIP_FLOW_TTL 10  // flow validity constraint (seconds from generation time)
#define EIP_IP_PROTO_NUM 253 // IP protocol number that indicates EIP
#define EIP_FLOW_TTL_THRESH 2 // maximum amount of time remaining on flow before type 3 is sent
//#define EIP_TADDRESS_VBYTES_THRESH 64000 // maximum bytes remaining on t-address before continuation is sent
#define EIP_TADDRESS_VBYTES_THRESH 1024000 // maximum bytes remaining on t-address before continuation is sent
#define EIP_TADDRESS_VTIME_THRESH 2 // maximum seconds remaining on t-address before continuation is sent
#define EIP_TADDRESS_CONT_BYTES_THRESH 27000 // number of bytes left on flow before type-1 is sent using continuation
// crypto stub constants
#define EIP_STUB_PUBKEY ""
#define EIP_STUB_PUBKEY_LEN 0
#define EIP_STUB_CERT ""
#define EIP_STUB_CERT_LEN 0
#define EIP_STUB_PRIVKEY ""
#define EIP_STUB_PRIVKEY_LEN 0
#define EIP_STUB_SIGLEN 128
// structure macros
#define EIPTYPE1 struct eip_type_1_datagram
#define EIPTYPE2 struct eip_type_2_datagram
#define EIPTYPE3 struct eip_type_3_datagram
#define EIPTYPE2C struct eip_type_2_datagram_cont
#define EIPTYPE3C struct eip_type_3_datagram_cont
#define TADDRESS struct eip_transient_address
// t-address field length macros
#define EIP_NET_ADDR_LEN 4     // IPv4 address length
#define EIP_TIMESTAMP_LEN 8    // TTL and timestamp field lengths
#define EIP_SIGNATURE_LEN 128  // 128 bytes = 1024-bit signature
#define EIP_PUBKEY_LEN 272     // 272 bytes = 1024-bit public key in PEM format
#define EIP_CERT_LEN 400       // 400 = 1024-bit public key in PEM format + signature
// eip datagram field length macros
#define EIP_TYPE_LEN 1         // EIP datagram type field length (always 1)
#define EIP_PROTO_LEN 1        // upper protocol field length (not used in current implementation)
#define EIP_DTOKEN_LEN 544     // delegation token length
#define EIP_FLOW_ID_LEN 16     // EIP flow ID length (flow ID is a 16-byte nonce)

// structure defining an eip transient address
TADDRESS {
  char src[EIP_NET_ADDR_LEN];   // network address of the source
  char dst[EIP_NET_ADDR_LEN];   // network address of the destination
  uint64_t vtime;               // expiry time for this address
  uint32_t vbytes;              // number of bytes constraint for this address
  char sig[EIP_SIGNATURE_LEN];  // signature field
  char pubkey[EIP_PUBKEY_LEN];  // public key of the issuer of this t-address (aka the destination)
  char cert[EIP_CERT_LEN];      // certificate for verifying the identity of the issuer
};
// structure defining eip type 1 datagram, which is used to establish new flows
EIPTYPE1 {
  char type[EIP_TYPE_LEN];      // type of this datagram
  char proto[EIP_PROTO_LEN];    // upper protocol byte
  TADDRESS src;                 // source t-address
  TADDRESS dst;                 // destination t-address
  char dtoken[EIP_DTOKEN_LEN];  // delegation token (see protocol spec)
  char flowid[EIP_FLOW_ID_LEN]; // flow id
  uint64_t flowttl;             // flow ttl (absolute time when this flow will expire)
  uint64_t timestamp;           // time when this datagram was created (to prevent replay)
  char sig[EIP_SIGNATURE_LEN];  // flowid + ttl + timestamp signature
  //uint8_t hasReplyT;            // source address present flag
  uint proc_time;           // processing time, stored on the packet
};
// structure defining eip type 2 datagram, which is used to re-use established flows
EIPTYPE2 {
  char type[EIP_TYPE_LEN];      // type of this datagram
  uint8_t continuation;         // continuation address present flag
  char proto[EIP_PROTO_LEN];    // upper protocol byte
  char flowid[EIP_FLOW_ID_LEN]; // flow id  
  uint proc_time;           // processing time, stored on the packet
};
// structure defining eip type 3 datagram, which is used to renew expiring flows
EIPTYPE3 {
  char type[EIP_TYPE_LEN];          // type of this datagram
  uint8_t continuation;             // continuation address present flag
  char proto[EIP_PROTO_LEN];        // upper protocol byte
  char old_flowid[EIP_FLOW_ID_LEN]; // old flow id    
  char new_flowid[EIP_FLOW_ID_LEN]; // new flow id    
  uint64_t flowttl;                 // new flow ttl
  uint64_t timestamp;               // time when this datagram was created (to prevent replay)  
  char sig[EIP_SIGNATURE_LEN];      // new flowid + ttl + timestamp signature
  uint proc_time;               // processing time, stored on the packet
};
// structure defining eip type 2 datagram, which is used to re-use established flows w/ continuation
EIPTYPE2C {
  char type[EIP_TYPE_LEN];      // type of this datagram
  uint8_t continuation;         // continuation address present flag
  char proto[EIP_PROTO_LEN];    // upper protocol byte
  char flowid[EIP_FLOW_ID_LEN]; // flow id  
  uint proc_time;               // processing time, stored on the packet
  TADDRESS cont_taddr;          // continuation t-address
};
// structure defining eip type 3 datagram, which is used to renew expiring flows w/ continuation
EIPTYPE3C {
  char type[EIP_TYPE_LEN];          // type of this datagram
  uint8_t continuation;             // continuation address present flag
  char proto[EIP_PROTO_LEN];        // upper protocol byte
  char old_flowid[EIP_FLOW_ID_LEN]; // old flow id    
  char new_flowid[EIP_FLOW_ID_LEN]; // new flow id    
  uint64_t flowttl;                 // new flow ttl
  uint64_t timestamp;               // time when this datagram was created (to prevent replay)
  char sig[EIP_SIGNATURE_LEN];      // new flowid + ttl + timestamp signature
  uint proc_time;               // processing time, stored on the packet
  TADDRESS cont_taddr;              // continuation t-address
};
// structure used for flow sign and verify operations
#define FLOW_SEAL struct eip_flow_seal_structure
FLOW_SEAL {
  char flowid[EIP_FLOW_ID_LEN]; // flowid
  char ttl[EIP_TIMESTAMP_LEN];                 // ttl field
  char ts[EIP_TIMESTAMP_LEN];               // timestamp field
};

#endif // ifndef __EIP_CONSTANTS_H__
