/**
 * File:
 *   eip-flowcache.cc
 * Description:
 *   Flow cache implementation using uthash
 *   Keys are a concatentation of src ip addr and dst ip addr
 * Author:
 *   ebj8
 * History:
 *   2011 10 06 ebj8: created for netfilter-based host implementation
 *   2012 08 01 ebj8: adapted for router module
 */
#ifndef __EIP_FLOWCACHE_CC__
#define __EIP_FLOWCACHE_CC__

#include <stdlib.h>
#include <stdio.h>
#include "./eip-flowcache.h"
#include "./memdmp.h"
#include "./eip-constants.h"
#include "./uthash.h"    // macro-based hash table implementation

const int debug_eip_flowcache = 0;

// *** global pointer to the hash table that stores the flow cache ***
FLOW_CACHE_RECORD *flowcache = NULL;

/*
 * add a record to the flow cache
 *  - takes a memory buffer as argument (must be eip type 1 datagram)
 */
void flowcache_add(EIPTYPE1* e1){
  // create and zero new record
  FLOW_CACHE_RECORD *rec = (FLOW_CACHE_RECORD*)malloc(sizeof(FLOW_CACHE_RECORD));
  memset(rec, 0, sizeof(FLOW_CACHE_RECORD));
  // copy data to new record
  if(debug_eip_flowcache){ printf("...caching flow for ");memdmp((char*)e1->dst.src, EIP_NET_ADDR_LEN*2);printf("\n"); }
  memcpy(&rec->key, e1->dst.src, EIP_NET_ADDR_LEN*2); // copy dstt src+dst net addr to key
  memcpy(rec->flowid, e1->flowid, EIP_FLOW_ID_LEN);   // copy flow id
  rec->ttl = e1->flowttl;
  rec->timestamp = e1->timestamp;
  rec->bytes = 0;
  memcpy(&rec->taddr, &e1->dst, sizeof(TADDRESS)); // copy destiantion t-address to this flow record for accounting and validation
  // debugging
  if(debug_eip_flowcache){ printf("...flow insert key: ");memdmp((char*)&rec->key, EIP_NET_ADDR_LEN*2);printf("\n");}// debug
  // printf("...flow insert id:  ");memdmp(rec->flowid, EIP_FLOW_ID_LEN);printf("\n");// debug  
  // insert new record into hash
  HASH_ADD(hh, flowcache, key, sizeof(FLOW_CACHE_KEY), rec);
}

void flowcache_update(FLOW_CACHE_RECORD* rec, EIPTYPE3 *e3){
  memcpy(rec->flowid, e3->new_flowid, EIP_FLOW_ID_LEN); // copy new flow id
  rec->ttl = e3->flowttl; // set new ttl
  rec->timestamp = e3->timestamp; // set generation timestamp
  // debugging
  // printf("...flow update key: ");memdmp((char*)&rec->key, EIP_NET_ADDR_LEN*2);printf("\n");// debug
  // printf("...flow update id:  ");memdmp(rec->flowid, EIP_FLOW_ID_LEN);printf("\n");// debug  
}

/*
 * delete a record from the cache (by pointer)
 * 
 */ 
void flowcache_del(FLOW_CACHE_RECORD *rec){
  HASH_DELETE(hh, flowcache, rec);
  free(rec);
}

/*
 * Returns a pointer to a flow cache record based on a lookup key
 * 
 */ 
FLOW_CACHE_RECORD* flowcache_lookup(FLOW_CACHE_KEY* needle){
  FLOW_CACHE_RECORD *rec;
  HASH_FIND(hh, flowcache, needle, sizeof(FLOW_CACHE_KEY), rec);
  if(debug_eip_flowcache){
	  printf("...flow lookup for key: ");memdmp((char*)needle, EIP_NET_ADDR_LEN*2);
	  printf((rec==NULL)?" FAIL":" OK");
	  printf("\n");
  }// debug
  return rec;
}

FLOW_CACHE_RECORD* flowcache_lookup_reverse(FLOW_CACHE_KEY* needle){
	FLOW_CACHE_KEY swapper;
	memcpy(&swapper, needle, sizeof(FLOW_CACHE_KEY));
	int tmp = swapper.src;
	swapper.src = swapper.dst;
	swapper.dst = tmp;
	return flowcache_lookup(&swapper);
}


/*
 * Returns a pointer to a flow cache record based on a flow id
 * 
 */ 
FLOW_CACHE_RECORD* flowcache_find(char* flowid){
  FLOW_CACHE_RECORD *flow = NULL;
  for(flow = flowcache; flow != NULL; flow = (FLOW_CACHE_RECORD*)flow->hh.next){
    if(memcmp(flowid, flow->flowid, EIP_FLOW_ID_LEN) == 0)
      return flow;
  }
  return flow;
}

#endif // ifndef __EIP_FLOWCACHE_CC__
