/**
 * File:
 *   eip-flowcache.h
 * Description:
 *   Flow cache implementation using uthash
 *   Keys are a concatentation of src ip addr and dst ip addr
 * Author:
 *   ebj8
 * History:
 *   2011 10 06 ebj8: created for netfilter-based host implementation
 *   2012 08 01 ebj8: adapted for router module
 */
#ifndef __EIP_FLOWCACHE_H__
#define __EIP_FLOWCACHE_H__

#include <stdlib.h>
#include <stdio.h>
#include "./memdmp.h"
#include "./eip-constants.h"
#include "./uthash.h"    // macro-based hash table implementation

#define FLOW_CACHE_KEY struct eip_flow_cache_key
FLOW_CACHE_KEY { // this structure will be our hash key
  int src; // network address of the source
  int dst; // network address of the destination
};


#define FLOW_CACHE_RECORD struct eip_flow_record
FLOW_CACHE_RECORD {
  FLOW_CACHE_KEY key; // hash key
  char flowid[EIP_FLOW_ID_LEN]; // flow id
  uint64_t ttl;       // time when this flow expires
  uint64_t timestamp; // time when the datagram that created this flow was created  
  uint32_t bytes;     // number of bytes that have traversed this flow
  TADDRESS taddr;     // t-address for the destination of this flow
  UT_hash_handle hh;  // makes the structure hashable
};

/**
 * add a record to the flow cache
 *  - takes a memory buffer as argument (must be eip type 1 datagram)
 */
void flowcache_add(EIPTYPE1* e1);

void flowcache_update(FLOW_CACHE_RECORD* rec, EIPTYPE3 *e3);

/**
 * delete a record from the cache (by pointer)
 * 
 */ 
void flowcache_del(FLOW_CACHE_RECORD *rec);

/**
 * Returns a pointer to a flow cache record based on a lookup key
 * 
 */ 
FLOW_CACHE_RECORD* flowcache_lookup(FLOW_CACHE_KEY* needle);

/**
 * Returns a pointer to a flow cache record based on a lookup key with the components reversed (swap src/dst)
 * 
 */ 
FLOW_CACHE_RECORD* flowcache_lookup_reverse(FLOW_CACHE_KEY* needle);


/**
 * Returns a pointer to a flow cache record based on a flow id
 * 
 */ 
FLOW_CACHE_RECORD* flowcache_find(char* flowid);


#endif // ifndef __EIP_FLOWCACHE_H__
