/**
 * File:
 *   eip-flowcache.cc
 * Description:
 *   Flow cache implementation using uthash
 *   Keys is flow id
 * Author:
 *   ebj8
 * History:
 *   2011 10 06 ebj8: created for netfilter-based host implementation
 *   2012 08 01 ebj8: adapted for router module
 *   2013 01 09 ebj8: forked this file from flowcache to add bitmap lookup
 */
#ifndef __EIP_ROUTERCACHE_CC__
#define __EIP_ROUTERCACHE_CC__

#include <stdlib.h>
#include <stdio.h>
#include "./eip-routercache.h"
#include "./memdmp.h"
#include "./eip-constants.h"
#include "./uthash.h"    // macro-based hash table implementation

const int debug_eip_routercache = 0;

// *** global pointer to the hash table that stores the flow cache ***
ROUTER_CACHE_RECORD * routercache = NULL;

// global bitmap of router records
ROUTER_CACHE_RECORD * routerbitmap[ROUTER_BITMAP_SIZE];

void routercache_init(){
	int i;
	for(i = 0; i < ROUTER_BITMAP_SIZE; i++){
		routerbitmap[i] = NULL;
	}
}

uint64_t routercache_get_bindex(char* flowid){
	if(debug_eip_routercache) printf("...routercache_get_bindex() called\n");
	uint64_t bindex = 0; // return value
	if(debug_eip_routercache){
		printf("...getting bindex for flowid: ");memdmp(flowid, EIP_FLOW_ID_LEN);printf("\n");
	}
	memcpy(&bindex, flowid, sizeof(bindex));
	bindex &= ROUTER_BITMAP_SIZE;
	if(debug_eip_routercache){
		printf("...                   bindex: %lu\n", bindex);
	}
	return bindex;
}

/*
 * add a record to the flow cache
 *  - takes a memory buffer as argument (must be eip type 1 datagram)
 */
void routercache_add(EIPTYPE1* e1){
  // create and zero new record
  ROUTER_CACHE_RECORD *rec = (ROUTER_CACHE_RECORD*)malloc(sizeof(ROUTER_CACHE_RECORD));
  memset(rec, 0, sizeof(ROUTER_CACHE_RECORD));
  // copy data to new record
  if(debug_eip_routercache){ printf("...caching flow for ");memdmp((char*)e1->flowid, EIP_FLOW_ID_LEN);printf("\n"); }
  memcpy(&rec->key, e1->flowid, EIP_FLOW_ID_LEN); // copy dstt src+dst net addr to key
  rec->ttl = e1->flowttl;
  rec->timestamp = e1->timestamp;
  rec->bytes = 0;
  memcpy(&rec->taddr, &e1->dst, sizeof(TADDRESS)); // copy destiantion t-address to this flow record for accounting and validation
  
  // copy vtime and vbytes so they are available with a single indirection
  rec->vbytes = e1->dst.vbytes;
  rec->vtime  = e1->dst.vtime;
  
  // debugging
  if(debug_eip_routercache){ printf("...flow insert key: ");memdmp((char*)&rec->key, EIP_FLOW_ID_LEN);printf("\n");}// debug
  // printf("...flow insert id:  ");memdmp(rec->flowid, EIP_FLOW_ID_LEN);printf("\n");// debug  
  // insert new record into hash
  uint64_t bindex = routercache_get_bindex(e1->flowid); // bitmap index
  if(debug_eip_routercache) printf("...routercache_add bindex: %lu\n", bindex);
  ROUTER_CACHE_RECORD *collision = routercache_lookup((ROUTER_CACHE_KEY*)e1->flowid);
  if(routerbitmap[bindex] == NULL && collision == NULL){
	routerbitmap[bindex] = rec;
  } else {
	HASH_ADD(hh, routercache, key, sizeof(ROUTER_CACHE_KEY), rec);
  }
}

void routercache_update(ROUTER_CACHE_RECORD* rec, EIPTYPE3 *e3){
  if(debug_eip_routercache) printf("...routercache_update() called\n");
  rec->ttl = e3->flowttl; // set new ttl
  rec->timestamp = e3->timestamp; // set generation timestamp
  uint64_t new_bindex = routercache_get_bindex(e3->new_flowid); // bitmap index
  uint64_t old_bindex = routercache_get_bindex(e3->old_flowid); // bitmap index
  if(debug_eip_routercache) printf("...routercache_change bindex: %lu > %lu\n", old_bindex, new_bindex);
  ROUTER_CACHE_RECORD *collision = routercache_lookup((ROUTER_CACHE_KEY*)e3->new_flowid);
  if(routerbitmap[new_bindex] == NULL && collision == NULL){
	// add this record directly to the bitmap in its new location
	routerbitmap[new_bindex] = rec;
	// remove the pointer to the prior version of the record
	if(routerbitmap[old_bindex] != NULL){
		routerbitmap[old_bindex] = NULL;
	} else {
		HASH_DELETE(hh, routercache, rec); // note, this does not free the memory for the record
	}
  } else {
	// place this record and the record with which it collides into the hash table
	HASH_ADD(hh, routercache, key, sizeof(ROUTER_CACHE_KEY), rec);
	HASH_ADD(hh, routercache, key, sizeof(ROUTER_CACHE_KEY), collision);
	routerbitmap[old_bindex] = NULL; // remove the pointer to the previous occupant of this bitmap location
  }
}

/*
 * delete a record from the cache (by pointer)
 * 
 */ 
void routercache_del(ROUTER_CACHE_RECORD *rec){
  if(rec == NULL) return;
  uint64_t bindex = routercache_get_bindex(rec->key.flowid); // bitmap index
  if(debug_eip_routercache) printf("...routercache_del bindex: %lu\n", bindex);
  if(routerbitmap[bindex] == NULL){
    HASH_DELETE(hh, routercache, rec);
  } else {
	  routerbitmap[bindex] == NULL;
  }
  free(rec);
}

/*
 * Returns a pointer to a flow cache record based on a lookup key
 * 
 */ 
ROUTER_CACHE_RECORD* routercache_lookup(ROUTER_CACHE_KEY* needle){
  uint64_t bindex = routercache_get_bindex(needle->flowid); // bitmap index
  if(debug_eip_routercache) printf("...routercache_lookup bindex: %lu\n", bindex);
  if(routerbitmap[bindex] != NULL){
	 if(debug_eip_routercache) printf("...routercache_lookup bitmap hit\n", bindex);
	 return routerbitmap[bindex]; 
  }
  ROUTER_CACHE_RECORD *rec; // return value
  HASH_FIND(hh, routercache, needle, sizeof(ROUTER_CACHE_KEY), rec);
  if(debug_eip_routercache) printf("...routercache_lookup bitmap miss\n", bindex);
  if(debug_eip_routercache){
	  printf("...flow lookup for key: ");memdmp((char*)needle, EIP_FLOW_ID_LEN);
	  printf((rec==NULL)?" FAIL":" OK");
	  printf("\n");
  }// debug
  return rec;
}

// deprecated
ROUTER_CACHE_RECORD* routercache_lookup_reverse(ROUTER_CACHE_KEY* needle){
/*
	ROUTER_CACHE_KEY swapper;
	memcpy(&swapper, needle, sizeof(ROUTER_CACHE_KEY));
	int tmp = swapper.src;
	swapper.src = swapper.dst;
	swapper.dst = tmp;
	return routercache_lookup(&swapper);
*/
return NULL;
}


/*
 * Returns a pointer to a flow cache record based on a flow id
 * 
 */ 
ROUTER_CACHE_RECORD* routercache_find(char* flowid){
  ROUTER_CACHE_RECORD *flow = NULL;
  for(flow = routercache; flow != NULL; flow = (ROUTER_CACHE_RECORD*)flow->hh.next){
    if(memcmp(flowid, &flow->key, EIP_FLOW_ID_LEN) == 0)
      return flow;
  }
  return flow;
}

#endif // ifndef __EIP_FLOWCACHE_CC__
