/**
 * File:
 *   eip-routercache.h
 * Description:
 *   Flow cache implementation using uthash
 *   Keys is concatentation flow id
 * Author:
 *   ebj8
 * History:
 *   2011 10 06 ebj8: created for netfilter-based host implementation
 *   2012 08 01 ebj8: adapted for router module
 *   2013 01 09 ebj8: forked this file from flowcache to add bitmap lookup
 */
#ifndef __EIP_ROUTERCACHE_H__
#define __EIP_ROUTERCACHE_H__

#include <stdlib.h>
#include <stdio.h>
#include "./memdmp.h"
#include "./eip-constants.h"
#include "./uthash.h"    // macro-based hash table implementation

#define ROUTER_BITMAP_SIZE 0xFFFFF

#define ROUTER_CACHE_KEY struct eip_router_cache_key
ROUTER_CACHE_KEY { // this structure will be our hash key
  char flowid[EIP_FLOW_ID_LEN]; // flow id
};

// in this implementation, we
// use an int for the key, 
// to represent a 4 byte IPv4 address
#define ROUTER_CACHE_RECORD struct eip_router_cache_record
ROUTER_CACHE_RECORD {
  ROUTER_CACHE_KEY key; // hash key  
  uint64_t ttl;         // time when this flow expires
  uint64_t timestamp;   // time when the datagram that created this flow was created  
  uint32_t bytes;       // number of bytes that have traversed this flow
  uint64_t vtime;       // taddr time constraint  
  uint32_t vbytes;      // taddr byte constraint
  TADDRESS taddr;       // t-address for the destination of this flow
  UT_hash_handle hh;    // makes the structure hashable
};

/**
 *  Initialize the pointers in the router bitmap
 */
void routercache_init();

/**
 * Get an index into the bitmap based on the flow id
 * 
 */ 
uint64_t routercache_get_bindex(char* flowid);

/**
 * add a record to the flow cache
 *  - takes a memory buffer as argument (must be eip type 1 datagram)
 */
void routercache_add(EIPTYPE1* e1);

void routercache_update(ROUTER_CACHE_RECORD* rec, EIPTYPE3 *e3);

/**
 * delete a record from the cache (by pointer)
 * 
 */ 
void routercache_del(ROUTER_CACHE_RECORD *rec);

/**
 * Returns a pointer to a flow cache record based on a lookup key
 * 
 */ 
ROUTER_CACHE_RECORD* routercache_lookup(ROUTER_CACHE_KEY* needle);

/**
 * Returns a pointer to a flow cache record based on a lookup key with the components reversed (swap src/dst)
 * 
 */ 
ROUTER_CACHE_RECORD* routercache_lookup_reverse(ROUTER_CACHE_KEY* needle);


/**
 * Returns a pointer to a flow cache record based on a flow id
 * 
 */ 
ROUTER_CACHE_RECORD* routercache_find(char* flowid);


#endif // ifndef __EIP_ROUTERCACHE_H__
