#ifndef __SIGN_VERIFY_CC__
#define __SIGN_VERIFY_CC__

#include "memdmp.h"

/*
 * File:
 *     sign-verify.cc
 * Description:
 *     Demonstration of RSA signing and verifying (all in-memory)
 * Compile:
 *     gcc -o rsa-sig-gen rsa-sig-gen.c -lssl
 * Run:
 *     ./rsa-sig-gen
 * Author:
 *     Eamon Johnson (ebj8@case.edu)
 * 
 * Notes:
 *    How to create a new key pair using command line openssl:
 *    - private key: openssl genrsa -out mykey.priv 1024
 *    - public key : openssl rsa -in mykey.priv -out mykey.pub -pubout
 *    How to encrypt and decrypt with command line openssl:
 *     - encrypt with public key : openssl rsautl -in foo.txt -out foo.enc -pubin -inkey mykey.pub -encrypt
 *     - decrypt with private key: openssl rsautl -in foo.enc -out foo.txt -inkey mykey.priv -decrypt
 *
 *    Note that in this example, the secret message is a null-terminated
 *    string.  If we want to sign an arbitrary buffer, then we must know its
 *    length and we must send that length to the signing functions instead of
 *    using strlen().
 *
 *    In this example, I have used a 1024-bit public key.
 *
 */

#include "eip-rsa-sign-verify.h"
#include <stdio.h>

// crypto globals
EVP_PKEY *eip_evp_puk = EVP_PKEY_new(); // allocate new envelope public key
EVP_PKEY *eip_evp_prk = EVP_PKEY_new(); // allocate new envelope private key
const EVP_MD *eip_md = EVP_sha1();      // get a pointer to the message digest implementation, could also use EVP_get_digestbyname("SHA1")
EVP_MD_CTX *eip_evp_md_ctx = EVP_MD_CTX_create();    // get a new envelope message digest context

const int debug_sign_verify = 0;

//function creates a new memory BIO and reads the RSA key
RSA *GetPublicKey(){
    char pubkeystring[strlen(PUBLIC_KEY_PEM)+1];
    strcpy(pubkeystring, PUBLIC_KEY_PEM);
    BIO *bp = BIO_new_mem_buf(pubkeystring, -1); // create new memory buffer BIO
    RSA *pubKey = PEM_read_bio_RSA_PUBKEY(bp,0,0,0);// read RSA key from BIO
    BIO_free(bp);
    return pubKey;
}

//function creates a new memory BIO and reads the RSA key
RSA *GetPublicKeyFromString(char* s, int len){
    char pubkeystring[len+1];
    strcpy(pubkeystring, s);
    BIO *bp = BIO_new_mem_buf(pubkeystring, -1); // create new memory buffer BIO
    RSA *pubKey = PEM_read_bio_RSA_PUBKEY(bp,0,0,0);// read RSA key from BIO
    BIO_free(bp);
    return pubKey;
}

//function creates a new memory BIO and reads the RSA key
RSA *GetPrivateKey(){
    char privkeystring[strlen(PRIVATE_KEY_PEM)+1];
    strcpy(privkeystring, PRIVATE_KEY_PEM);
    BIO *bp = BIO_new_mem_buf(privkeystring, -1); // create new memory buffer BIO
    RSA *privKey = PEM_read_bio_RSAPrivateKey(bp,0,0,0);// read RSA key from BIO
    BIO_free(bp);
    return privKey;
}



void eip_rsa_sign(char* sig, char* data, unsigned int datalen){
  // get the keys
  RSA *privKey = GetPrivateKey();           // get pointer to private key
  EVP_PKEY_set1_RSA(eip_evp_prk, privKey);  // copy the envelope private key from the RSA private key
  unsigned int sig_len = 0;                 // holds the length of the generated signature
  //unsigned char sig[EVP_PKEY_size(eprk)];   // allocate buffer for signature
  //printf("Maximum signature size in bytes is %d\n", EVP_PKEY_size(eprk));//debug

  // do the signing
  EVP_SignInit(eip_evp_md_ctx, eip_md);                    // initialize the signature context with our selected hash alg (default implementation);    
  EVP_SignUpdate(eip_evp_md_ctx, (unsigned char*)data, datalen);    // add the message data to the hash
  EVP_SignFinal(eip_evp_md_ctx, (unsigned char*)sig, &sig_len, eip_evp_prk);  // sign the hash with the private key and place the signature in "sig"
  if(debug_sign_verify){printf("...   signed: ");memdmp(data, datalen);printf("\n");} //debug
}

int eip_rsa_verify(char* data, unsigned int datalen, char* sig, unsigned int siglen, char* pubkey, int pubkeylen){
  // get the keys
  RSA *pubKey = GetPublicKeyFromString(pubkey, pubkeylen);           // get pointer to public key
  EVP_PKEY_set1_RSA(eip_evp_puk, pubKey);  // copy the envelope public key from the RSA public key
  // do the verification
  int v = 0;
  unsigned int sig_len = 128;                
  EVP_VerifyInit(eip_evp_md_ctx, eip_md);                  // intialize the message digest context for verification using our selected hash alg    
  EVP_VerifyUpdate(eip_evp_md_ctx, data, datalen);  // add message data to the hash
  v = EVP_VerifyFinal(eip_evp_md_ctx,(unsigned char*)sig,sig_len,eip_evp_puk);// verify the signature
  if(debug_sign_verify){printf("\n...verifying: ");memdmp(data, datalen);printf(" ");}//debug
  if(debug_sign_verify){printf((v?"\n...signature verified ":"\n...signature is not correct "));}//debug

  return v;
}

#define DATALEN 20

// main function: demonstrates use of EVP API to sign and verify signatures
int sign_verify_demo(void){
    ///////////////
    // data to sign: src ip 10.1.1.2, dst ip 10.2.2.2, validity constraint = 12 bytes of 255;
    unsigned char b[DATALEN];
    b[0] = 10;
    b[1] = 1;
    b[2] = 1;
    b[3] = 2;
    b[4] = 10;
    b[5] = 2;
    b[6] = 2;
    b[7] = 2;
    b[8] = 255;
    b[9] = 255;
    b[10] = 255;
    b[11] = 255;
    b[12] = 255;
    b[13] = 255;
    b[14] = 255;
    b[15] = 255;
    b[16] = 255;
    b[17] = 255;
    b[18] = 255;
    b[19] = 255;
    //////////////////////////

    OpenSSL_add_all_digests();                // load all known digests
    unsigned char *msg = b;      // create pointer to our secret message
    RSA *pubKey = GetPublicKey();             // get pointer to public key   
    RSA *privKey = GetPrivateKey();           // get pointer to private key
    EVP_PKEY *epuk = EVP_PKEY_new();          // allocate new envelope public key
    EVP_PKEY *eprk = EVP_PKEY_new();          // allocate new envelope private key
    EVP_PKEY_set1_RSA(epuk, pubKey);          // copy the envelope public key from the RSA public key
    EVP_PKEY_set1_RSA(eprk, privKey);         // copy the envelope private key from the RSA private key
    EVP_MD_CTX *ctx = EVP_MD_CTX_create();    // get a new envelope message digest context
    EVP_MD_CTX_init(ctx);                     // initialize envelope message digest context
    const EVP_MD *md = EVP_sha1();            // get a pointer to the message digest implementation, could also use EVP_get_digestbyname("SHA1")
    unsigned int sig_len = 0;                 // holds the length of the generated signature
    unsigned char sig[EVP_PKEY_size(eprk)];   // allocate buffer for signature
    printf("Maximum signature size in bytes is %d\n", EVP_PKEY_size(eprk));//debug

    // do the signing
    EVP_SignInit(ctx, md);                    // initialize the signature context with our selected hash alg (default implementation);    
    EVP_SignUpdate(ctx, msg, DATALEN);    // add the message data to the hash
    EVP_SignFinal(ctx, sig, &sig_len, eprk);  // sign the hash with the private key and place the signature in "sig"
    printf("Created signature of length %d\n", sig_len); //debug

    // verify the signature
    int v = 0;                                // verification status indicator
    EVP_VerifyInit(ctx, md);                  // intialize the message digest context for verification using our selected hash alg    
    EVP_VerifyUpdate(ctx, msg, DATALEN);  // add message data to the hash
    v = EVP_VerifyFinal(ctx,sig,sig_len,epuk);// verify the signature
    printf((v?"Signature verified\n":"Signature is not correct\n"));//debug

    //put the signature in a file
    FILE *f = fopen("somebytes.sig", "w");
    fwrite(sig, sizeof(char), sig_len, f);
    fclose(f);

    // clean up and exit
    EVP_MD_CTX_destroy(ctx);                  // destroy the envelope message digest context
    EVP_PKEY_free(epuk);                      // release envelope public key
    EVP_PKEY_free(eprk);                      // release envelope private key
    RSA_free(pubKey);                         // release public key memory
    RSA_free(privKey);                        // release private key memory
    return EXIT_SUCCESS;
}

#endif // ifndef __SIGN_VERIFY_CC__
