#ifndef __SIGN_VERIFY_H__
#define __SIGN_VERIFY_H__

#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/rsa.h>
#include <string.h>
#include "eip-crypto.h"
#include "memdmp.h"

/*
 * File:
 *     sign-verify.h
 * Description:
 *     Demonstration of RSA signing and verifying (all in-memory)
 *      - adapted from an example I found that was directed more toward SSL on Windows
 *      - http://nlnetlabs.nl/downloads/publications/hsm/hsm.html is also a good resource for the Open SSL EVP (EnVeloPe) API
 * Compile:
 *     gcc -o rsa-sig-gen rsa-sig-gen.c -lssl
 * Run:
 *     ./rsa-sig-gen
 * Author:
 *     Eamon Johnson (ebj8@case.edu)
 * 
 * Notes:
 *    How to create a new key pair using command line openssl:
 *    - private key: openssl genrsa -out mykey.priv 1024
 *    - public key : openssl rsa -in mykey.priv -out mykey.pub -pubout
 *    How to encrypt and decrypt with command line openssl:
 *     - encrypt with public key : openssl rsautl -in foo.txt -out foo.enc -pubin -inkey mykey.pub -encrypt
 *     - decrypt with private key: openssl rsautl -in foo.enc -out foo.txt -inkey mykey.priv -decrypt
 *
 *    Note that in this example, the secret message is a null-terminated
 *    string.  If we want to sign an arbitrary buffer, then we must know its
 *    length and we must send that length to the signing functions instead of
 *    using strlen().
 *
 *    In this example, I have used a 1024-bit public key.
 *
 */

#define SECRET_MESSAGE \
"SIGNATU"

#define PUBLIC_KEY_PEM \
"-----BEGIN PUBLIC KEY-----\n" \
"MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQConOIlN0sVntjwQJL5c/Z9aXrM\n" \
"faYVWBHrbxK80BlGP3zhK9yz2ETrrMxu2WvwkGmxRdg2/fTou746Qxj+i2fHGbxo\n" \
"CeTEswwPRaz4NWKInaEtVmZ6iPWIIf4zIYg1soxOGR/EZAeyKXg507n1wFMM4oug\n" \
"Sgt7FE/WHn9kXP+8nwIDAQAB\n" \
"-----END PUBLIC KEY-----"

#define PRIVATE_KEY_PEM \
"-----BEGIN RSA PRIVATE KEY-----\n" \
"MIICXAIBAAKBgQConOIlN0sVntjwQJL5c/Z9aXrMfaYVWBHrbxK80BlGP3zhK9yz\n" \
"2ETrrMxu2WvwkGmxRdg2/fTou746Qxj+i2fHGbxoCeTEswwPRaz4NWKInaEtVmZ6\n" \
"iPWIIf4zIYg1soxOGR/EZAeyKXg507n1wFMM4ougSgt7FE/WHn9kXP+8nwIDAQAB\n" \
"AoGAWM1nfVpFvTlP4BjiT2b3ps1q01BJrGwZ4/k30SF4bjr5tk23izCX+9eiQ69W\n" \
"3ffw8mBqo2vVFM3sPhsgis849BSCYRaPiL/qFu27Ke6KAvOREj/MSusvsb6aO3Br\n" \
"UKtGytpv9kdXUSiGG6St1ZmzmrjyqBqYU0ePGSLnm1iPl2ECQQDT9/Q0E6gDlRw7\n" \
"F4LC5UZwkzJ0lfhD2JHw1EZzoNKu3DteKtZ/KQdYoQ9Ylt6AVmvqjMc6JR+F7zvG\n" \
"98Q7+HTtAkEAy6NcihJES6R3vrVxVyqBS+iV6t+1RahNdN1tRJBsbe5FGK9TAXUT\n" \
"x7zwT+4tS9hId37qDV/65o+XRXTysTCyOwJABzCe8PFFvrFdHsgCYHgEdqcXjArW\n" \
"O/7pnGCm7x8iZloEDYKoefG7VQ9JhPjBHVtRF8ikrV06/4EI5wvadSpNVQJBAMaN\n" \
"NgF55hVK+2Z7S1OOIht9YA9uqDSO8VbtP0kG+HIHTF/RMsHNAchStI160beN3rqE\n" \
"iHkaCG0dvIUAjlGuBVkCQF+ohO430Gzs1ZOtyYCljFXKoA0O3qx7VBvx/CefiEm5\n" \
"x1atsxysbZxxp5XAOjpcAOttFfsXhlBOz5V7MzdllfI=\n" \
"-----END RSA PRIVATE KEY-----"


//function creates a new memory BIO and reads the RSA key
RSA *GetPublicKey();

//function creates a new memory BIO and reads the RSA key
RSA *GetPublicKeyFromString(char* s, int len);

//function creates a new memory BIO and reads the RSA key
RSA *GetPrivateKey();

void eip_rsa_sign(char* sig, char* data, unsigned int datalen);

int eip_rsa_verify(char* data, unsigned int datalen, char* sig, unsigned int siglen, char* pubkey, int pubkeylen);

#define DATALEN 20

// main function: demonstrates use of EVP API to sign and verify signatures
int sign_verify_demo(void);

#endif // ifndef __SIGN_VERIFY_H__
