/*
 *  File: eip_timer.cc
 */
#ifndef __EIP_TIMER_CC__
#define __EIP_TIMER_CC__

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "eip_timer.h"

// macro for selection of the clock to use when calculating router performance
//#define EIP_ROUTER_CLOCK CLOCK_PROCESS_CPUTIME_ID
#define EIP_ROUTER_CLOCK CLOCK_REALTIME
//#define EIP_ROUTER_CLOCK CLOCK_MONOTONIC
//#define EIP_ROUTER_CLOCK CLOCK_THREAD_CPUTIME_ID


// clock timing globals
timespec eip_timer_starttime;
timespec eip_timer_stoptime;
unsigned int eip_timer_nanodiff;

#ifdef EIP_PAPI_CYCLE_COUNTING
// PAPI cycle timing globals
long long eip_cyc_start;
long long eip_cyc_stop;
long long eip_cyc_diff;  
#endif

#ifdef EIP_ASM_CYCLE_COUNTING
/* Initialize the cycle counter */
static unsigned asm_cyc_hi = 0;
static unsigned asm_cyc_lo = 0;
static double asm_cyc_diff = 0;
#endif

const int debug_eip_timer = 1;

void eip_timer_start(){
	#ifdef EIP_DISABLE_ROUTER_PERF_ASSESS
	return;
	#endif 
	clock_gettime(EIP_ROUTER_CLOCK, &eip_timer_starttime);
	#ifdef EIP_PAPI_CYCLE_COUNTING
	eip_cyc_start = PAPI_get_real_cyc();
	#endif
	#ifdef EIP_ASM_CYCLE_COUNTING
	eip_asm_start_counter();
	#endif
}

void eip_timer_stop(unsigned int* elapsed_nanoseconds){
	#ifdef EIP_DISABLE_ROUTER_PERF_ASSESS
	return;
	#endif 
	#ifdef EIP_ASM_CYCLE_COUNTING
	asm_cyc_diff = eip_asm_get_counter();
	#endif
	#ifdef EIP_PAPI_CYCLE_COUNTING
	eip_cyc_stop = PAPI_get_real_cyc();
	eip_cyc_diff = eip_cyc_stop - eip_cyc_start;
	#endif
	clock_gettime(EIP_ROUTER_CLOCK, &eip_timer_stoptime);
    if(eip_timer_stoptime.tv_sec > eip_timer_starttime.tv_sec){
		eip_timer_stoptime.tv_nsec += 1000000000 * (eip_timer_stoptime.tv_sec - eip_timer_starttime.tv_sec);
    }
    eip_timer_nanodiff = eip_timer_stoptime.tv_nsec - eip_timer_starttime.tv_nsec;
    if(debug_eip_timer){
		printf("NSEC CLOCK PERF: %u\n", eip_timer_nanodiff);
		#ifdef EIP_PAPI_CYCLE_COUNTING    
		printf("PAPI CYCLE PERF: %lld\n", eip_cyc_diff);
		#endif
		#ifdef EIP_ASM_CYCLE_COUNTING
		printf("IA32 CYCLE PERF: %g\n", asm_cyc_diff);
		#endif
	}
	memcpy(elapsed_nanoseconds, &eip_timer_nanodiff, sizeof(unsigned int));
}

#ifdef EIP_ASM_CYCLE_COUNTING

/* Set *hi and *lo to the high and low order bits of the cycle counter.
   Implementation requires assembly code to use the rdtsc instruction. */
void eip_asm_access_counter(unsigned *hi, unsigned *lo){
  asm("rdtscp; movl %%edx,%0; movl %%eax,%1" /* Read cycle counter */
  : "=r" (*hi), "=r" (*lo) /* and move results to */
  : /* No input */ /* the two outputs */
  : "%edx", "%eax");
}

/* Record the current value of the cycle counter. */
void eip_asm_start_counter(){
  eip_asm_access_counter(&asm_cyc_hi, &asm_cyc_lo);
}

/* Return the number of cycles since the last call to start_counter. */
double eip_asm_get_counter(){
  unsigned ncyc_hi, ncyc_lo;
  unsigned hi, lo, borrow;
  double result;
  /* Get cycle counter */
  eip_asm_access_counter(&ncyc_hi, &ncyc_lo);
  /* Do double precision subtraction */
  lo = ncyc_lo - asm_cyc_lo;
  borrow = lo > ncyc_lo;
  hi = ncyc_hi - asm_cyc_hi - borrow;
  result = (double) hi * (1 << 30) * 4 + lo;
  if (result < 0) {
    fprintf(stderr, "Error: counter returns neg value: %.0f\n", result);
  }
  return result;
}
#endif // EIP_ASM_CYCLE_COUNTING

#endif // __EIP_TIMER_CC__
