/**
 * File:
 *   eip-signcache.cc
 * Description:
 *   Signed taddr cache implementation using uthash
 *   Keys are a concatentation of src ip addr and dst ip addr
 * Author:
 *   ebj8
 * History:
 *   2011 12 28 ebj8: created for netfilter-based host implementation
 */
#ifndef __EIP_SIGNCACHE_CC__
#define __EIP_SIGNCACHE_CC__

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "./eip-signcache.h"
#include "../common/memdmp.h"
#include "../common/eip-constants.h"
#include "../common/uthash.h"    // macro-based hash table implementation
#include "../common/eip-rsa-sign-verify.h"
#include "../common/eip-crypto.h"

const int debug_eip_signcache = 1;

// *** global pointer to the hash table that stores the flow cache ***
SIGN_CACHE_RECORD *signcache = NULL;

// global thread pointer
pthread_t signerThread;
// global mutex for controlling access to this cache
pthread_mutex_t signerMutex;
// global flag for turn-taking.  0=main thread's turn; 1=signer's thread's turn
int signerTurn = 0;

void *signcache_run(void *arg){
  if(debug_eip_signcache) printf("SIGNER START\n");
  while(1){
	while(signerTurn == 0); // spin while waiting for turn
    pthread_mutex_lock(&signerMutex);
    signcache_sweep();
    pthread_mutex_unlock(&signerMutex);
    signerTurn = 0;
  }
}

int signcache_get(iphdr* ip, TADDRESS* taddr){
	int rval = 0; // return value
	SIGN_CACHE_RECORD *scr = signcache_lookup((SIGN_CACHE_KEY*)(&ip->saddr));
	while(signerTurn == 1); // spin while waiting for turn
	pthread_mutex_lock(&signerMutex);
	if(scr == NULL){
		printf("SIGNER INSERT: signature requested for new pair\n");		
		signcache_insert(ip, taddr); // insert this key into the cache for later asynchronous signing
		rval = 0; // return for synchronous signing
	} else {
		if(scr->ready == 1){
			printf("SIGNER READY: signature ready when requested\n");			
			memcpy(&scr->taddr, taddr, sizeof(TADDRESS)); // copy ready signature to taddr
			scr->ready = 0;
			rval = 1;
		} else {
			printf("SIGNER FAULT: singature not ready when requested\n");
			rval = 0; // return for synchronous signing
		}
	}
	pthread_mutex_unlock(&signerMutex);
	signerTurn = 1;
	return rval;
}

void signcache_insert(iphdr* ip, TADDRESS* taddr){
// create and zero new record
  SIGN_CACHE_RECORD *rec = (SIGN_CACHE_RECORD*)malloc(sizeof(SIGN_CACHE_RECORD));
  memset(rec, 0, sizeof(SIGN_CACHE_RECORD));
  // copy data to new record
  rec->ready = 0; // indicates this record needs a new signature
  memcpy(&rec->taddr, taddr, sizeof(TADDRESS));
  memcpy(&rec->key, &ip->saddr, EIP_NET_ADDR_LEN*2); // copy ip src+dst net addr to key
  if(debug_eip_signcache){ printf("SIGNER insert key: ");memdmp((char*)&rec->key, EIP_NET_ADDR_LEN*2);printf("\n");}// debug  
  // insert new record into hash
  HASH_ADD(hh, signcache, key, sizeof(SIGN_CACHE_KEY), rec);
}

void signcache_sweep(){
  SIGN_CACHE_RECORD *scr = NULL;
  for(scr = signcache; scr != NULL; scr = (SIGN_CACHE_RECORD*)scr->hh.next){
	if(scr->ready == 0){
      signcache_sign(&scr->taddr);
      scr->ready = 1;
	}
  }
}

void signcache_sign(TADDRESS* taddr){
  //return; // uncomment to exclude signature operation
  if(debug_eip_signcache) printf("SIGNER ASYNC: signing\n");//debug
  const int datalen = EIP_NET_ADDR_LEN*2 + sizeof(uint64_t) + sizeof(uint32_t);
  char data[datalen]; // allocate buffer for data to be signed
  memcpy(data, &taddr->src, EIP_NET_ADDR_LEN*2); // copy src and dst addresses (contiguous memory)
  memcpy(data + EIP_NET_ADDR_LEN*2, &taddr->vtime, sizeof(uint64_t));                    // copy vtime to buffer
  memcpy(data + EIP_NET_ADDR_LEN*2 + sizeof(uint64_t), &taddr->vbytes, sizeof(uint32_t)); // copy vbytes to buffer
  eip_rsa_sign(taddr->sig, data, datalen); 
}

void signcache_setup(){
	if(debug_eip_signcache) printf("SIGNER SETUP\n");
	pthread_attr_t attr;
	pthread_mutex_init(&signerMutex, NULL);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	pthread_create(&signerThread, &attr, signcache_run, (void*)NULL);
	pthread_attr_destroy(&attr);
}

///////////////////////////////////////////////////////

/*
 * add a record to the flow cache
 *  - takes a memory buffer as argument (must be eip type 1 datagram)
 */
void signcache_add(EIPTYPE1* e1){
  /*
  
  
  
  // create and zero new record
  FLOW_CACHE_RECORD *rec = (FLOW_CACHE_RECORD*)malloc(sizeof(FLOW_CACHE_RECORD));
  memset(rec, 0, sizeof(FLOW_CACHE_RECORD));
  // copy data to new record
  if(debug_eip_flowcache){ printf("...caching flow for ");memdmp((char*)e1->dst.src, EIP_NET_ADDR_LEN*2);printf("\n"); }
  memcpy(&rec->key, e1->dst.src, EIP_NET_ADDR_LEN*2); // copy dstt src+dst net addr to key
  memcpy(rec->flowid, e1->flowid, EIP_FLOW_ID_LEN);   // copy flow id
  rec->ttl = e1->flowttl;
  rec->timestamp = e1->timestamp;
  rec->bytes = 0;
  memcpy(&rec->taddr, &e1->dst, sizeof(TADDRESS)); // copy destiantion t-address to this flow record for accounting and validation
  // debugging
  if(debug_eip_flowcache){ printf("...flow insert key: ");memdmp((char*)&rec->key, EIP_NET_ADDR_LEN*2);printf("\n");}// debug
  // printf("...flow insert id:  ");memdmp(rec->flowid, EIP_FLOW_ID_LEN);printf("\n");// debug  
  // insert new record into hash
  HASH_ADD(hh, flowcache, key, sizeof(FLOW_CACHE_KEY), rec);
  */ 
}

void signcache_update(SIGN_CACHE_RECORD* rec, EIPTYPE3 *e3){
  /*
  memcpy(rec->flowid, e3->new_flowid, EIP_FLOW_ID_LEN); // copy new flow id
  rec->ttl = e3->flowttl; // set new ttl
  rec->timestamp = e3->timestamp; // set generation timestamp
  // debugging
  // printf("...flow update key: ");memdmp((char*)&rec->key, EIP_NET_ADDR_LEN*2);printf("\n");// debug
  // printf("...flow update id:  ");memdmp(rec->flowid, EIP_FLOW_ID_LEN);printf("\n");// debug  
  * */
}

/*
 * delete a record from the cache (by pointer)
 * 
 */ 
void signcache_del(SIGN_CACHE_RECORD *rec){
  HASH_DELETE(hh, signcache, rec);
  free(rec);
}

/*
 * Returns a pointer to a flow cache record based on a lookup key
 * 
 */ 
SIGN_CACHE_RECORD* signcache_lookup(SIGN_CACHE_KEY* needle){
  SIGN_CACHE_RECORD *rec;
  HASH_FIND(hh, signcache, needle, sizeof(SIGN_CACHE_KEY), rec);
  if(debug_eip_signcache){
	  printf("...esign lookup for key: ");memdmp((char*)needle, EIP_NET_ADDR_LEN*2);
	  printf((rec==NULL)?" FAIL":" OK");
	  printf("\n");
  }// debug
  return rec;
}

SIGN_CACHE_RECORD* signcache_lookup_reverse(SIGN_CACHE_KEY* needle){
	/*
	FLOW_CACHE_KEY swapper;
	memcpy(&swapper, needle, sizeof(FLOW_CACHE_KEY));
	int tmp = swapper.src;
	swapper.src = swapper.dst;
	swapper.dst = tmp;
	return flowcache_lookup(&swapper);
	* */
	return NULL;
}


/*
 * Returns a pointer to a flow cache record based on a flow id
 * 
 */ 
SIGN_CACHE_RECORD* signcache_find(char* signid){/*
  FLOW_CACHE_RECORD *flow = NULL;
  for(flow = flowcache; flow != NULL; flow = (FLOW_CACHE_RECORD*)flow->hh.next){
    if(memcmp(flowid, flow->flowid, EIP_FLOW_ID_LEN) == 0)
      return flow;
  }
  return flow;
  */
  return NULL;
}

#endif // ifndef __EIP_SIGNCACHE_CC__
