/**
 * eip_host.cc
 * 
 * Implementation of EIP host processing.
 * 
 */ 

#include "eip_host.h"
#include "../common/eip_timer.h"


const int debug_eip_host = 0;     // debug flag
const int debug_eip_host_in = 1;  // debug flag (inbound)
const int debug_eip_host_out = 1; // debug flag (outbound)

// global pointers for working with EIP datagrams
EIPTYPE1 *e1;
EIPTYPE2 *e2;
EIPTYPE2C *e2c;
EIPTYPE3 *e3;
EIPTYPE3C *e3c;

/**
 * Validate the EIP headers of an inbound packet.
 * @param pkt  Pointer to packet data buffer containing a IP packet with appended EIP headers.
 * @param len  Length of the data in the buffer.
 * @return  EIP_VALID if the packet is valid, EIP_INVALID otherwise.
 * 
 */ 
int eip_host_validate(const unsigned char *pkt, const unsigned int len){
	uint16_t originalTotalLength;
	memcpy(&originalTotalLength, pkt + len - 2, 2); // get original length from end of packet
	uint16_t originalSize = ntohs(originalTotalLength);
	unsigned char* eip_header = (unsigned char*)(pkt + originalSize);
	if(debug_eip_host) printf("DEBUG: inbound eip type = %d\n", eip_header[0]);//debug
	switch(eip_header[0]){ // first byte of extended packet is EIP type
		case EIP_TYPE_SETUP:
			if(eip_validate_type1(eip_header, pkt)){
				if(debug_eip_host) printf("DEBUG: TYPE 1 valid\n");//debug
				return EIP_VALID;
			} else {
				if(debug_eip_host) printf("DEBUG: TYPE 1 INVALID\n");//debug
				return EIP_INVALID;
			}
		    break;
		case EIP_TYPE_REUSE:
			e2 = (EIPTYPE2*) eip_header;
			if(e2->continuation){
				if(debug_eip_host) printf("DEBUG: type 2 has inbound continuation\n");//debug
				if(eip_validate_type2(eip_header, pkt, EIP_CONT_YES)){
					if(debug_eip_host) printf("DEBUG: TYPE 2c valid\n");//debug
					return EIP_VALID;
				} else {
					if(debug_eip_host) printf("DEBUG: TYPE 2c INVALID\n");//debug
					return EIP_INVALID;
				}
			} else {
				if(eip_validate_type2(eip_header, pkt, EIP_CONT_NO)){
					if(debug_eip_host) printf("DEBUG: TYPE 2 valid\n");//debug
					return EIP_VALID;					
				} else {
					if(debug_eip_host) printf("DEBUG: TYPE 2 INVALID\n");//debug
					return EIP_INVALID;
				}
			}
			return EIP_INVALID;
		    break;
		case EIP_TYPE_RENEW:
			e3 = (EIPTYPE3*) eip_header;
			if(e3->continuation){
				if(debug_eip_host) printf("DEBUG: type 3 has inbound continuation\n");//debug
				if(eip_validate_type3(eip_header, pkt, EIP_CONT_YES)){
					if(debug_eip_host) printf("DEBUG: TYPE 3c valid\n");//debug
					return EIP_VALID;
				} else {
					if(debug_eip_host) printf("DEBUG: TYPE 3c INVALID\n");//debug
					return EIP_INVALID;
				}
			} else {
				if(eip_validate_type3(eip_header, pkt, EIP_CONT_NO)){
					if(debug_eip_host) printf("DEBUG: TYPE 3 valid\n");//debug
					return EIP_VALID;
				} else {
					if(debug_eip_host) printf("DEBUG: TYPE 3 INVALID\n");//debug
					return EIP_INVALID;
				}
			}
		    break;
	}
	if(debug_eip_host) printf("DEBUG: ***INVALID PACKET***\n");//debug
	return EIP_INVALID; // unreachable
}

/**
 * Creates new EIP headers for an outbound packet.
 * @param headers  Buffer where new EIP header data will be written.
 * @param pkt      Buffer containing original packet.
 * @param len      Length of original packet data.
 * @return         Length of new EIP header data.
 * 
 */ 
unsigned int eip_get_new_headers(unsigned char *headers, const unsigned char *pkt, const unsigned int len){
	unsigned int header_length = 0; // return value
	FLOW_CACHE_RECORD *flow = NULL;	
	FLOW_CACHE_RECORD *reverseFlow = NULL;	
	ADDR_CACHE_RECORD *addr = NULL;
	// check to see if flow is in cache
	flow = flowcache_lookup((FLOW_CACHE_KEY*)&((iphdr*)pkt)->saddr); //the key struct is the ip saddr and daddr, which start at byte 12
	reverseFlow = flowcache_lookup_reverse((FLOW_CACHE_KEY*)&((iphdr*)pkt)->saddr); //the key struct is the ip saddr and daddr, which start at byte 12
	if(flow){
		bug("eip_get_new_headers:");addr = addrcache_lookup((ADDR_CACHE_KEY*)&((iphdr*)pkt)->saddr);
		if(addr && addr->sent==0 && eip_use_continuation(flow)){
			// SEND TYPE1 (from continuation)
			header_length = sizeof(EIPTYPE1)+NF_OVERHEAD_BYTES;
			eip_send_type1(headers, pkt, len, EIP_CONT_YES);
			if(debug_eip_host) printf("DEBUG: outbound eip type = 1 (from continuation)\n");//debug
		} else {
			bug("eip_get_new_headers:");addr = addrcache_lookup_reverse((ADDR_CACHE_KEY*)&((iphdr*)pkt)->saddr);
			if(eip_flow_isValid(flow)){ // flow is valid -- okay to reuse flow
				if(eip_capability_isValid(reverseFlow)){ // taddr is valid
					// SEND TYPE 2
					header_length = sizeof(EIPTYPE2)+NF_OVERHEAD_BYTES;
					eip_send_type2(headers, pkt, len, flow, EIP_CONT_NO);
					if(debug_eip_host) printf("DEBUG: outbound eip type = 2\n");//debug
				} else { // taddr is invalid
					if(addr && addr->sent==1){ // continuation has already been sent
						// SEND TYPE 2
						header_length = sizeof(EIPTYPE2)+NF_OVERHEAD_BYTES;
						eip_send_type2(headers, pkt, len, flow, EIP_CONT_NO);
						if(debug_eip_host) printf("DEBUG: outbound eip type = 2\n");//debug
					} else { // continuation has not been sent
						// SEND TYPE 2 W CONTINUATION
						if(addr) if(!addr->sent) if(debug_eip_host) printf("DEBUG: addr->sent == 0\n");//debug
						if(addr) if(addr->sent) if(debug_eip_host) printf("DEBUG: addr->sent == 1\n");//debug
						header_length = sizeof(EIPTYPE2C)+NF_OVERHEAD_BYTES;
						eip_send_type2(headers, pkt, len, flow, EIP_CONT_YES);
						if(debug_eip_host) printf("DEBUG: outbound eip type = 2 (with continuation)\n");//debug					
					}
				}				
			} else { // flow is invalid -- need to renew flow
				if(eip_capability_isValid(reverseFlow)){ // taddr is valid
					// SEND TYPE 3
					header_length = sizeof(EIPTYPE3)+NF_OVERHEAD_BYTES;
					eip_send_type3(headers, pkt, len, flow, EIP_CONT_NO);
					if(debug_eip_host) printf("DEBUG: outbound eip type = 3\n");//debug
				} else {
					if(addr && addr->sent){ // continuation has already been sent
						// SEND TYPE 3
						header_length = sizeof(EIPTYPE3)+NF_OVERHEAD_BYTES;
						eip_send_type3(headers, pkt, len, flow, EIP_CONT_NO);
						if(debug_eip_host) printf("DEBUG: outbound eip type = 3\n");//debug
					} else { // continuation has not been sent
						// SEND TYPE 3 W CONTINUATION
						header_length = sizeof(EIPTYPE3C)+NF_OVERHEAD_BYTES;
						eip_send_type3(headers, pkt, len, flow, EIP_CONT_YES);
						if(debug_eip_host) printf("DEBUG: outbound eip type = 3 (with continuation)\n");//debug
					}
				}
			}
		}
	} else {
		// SEND TYPE 1 (new)
		header_length = sizeof(EIPTYPE1)+NF_OVERHEAD_BYTES;
		eip_send_type1(headers, pkt, len, EIP_CONT_NO);
		if(debug_eip_host) printf("DEBUG: outbound eip type = 1\n");//debug
	}
	// update flow accounting with bytes of new headers + bytes of original packet
	flow->bytes += header_length + len;
	// return length of new EIP headers
	return header_length;
}

void eip_nonce(char* buf, const int len){
  unsigned int i = 0;  
  for(; i<len; i++) buf[i] = rand() % (BYTE + 1);
}

//
// validation functions
//
int eip_validate_type1(const unsigned char* eip, const unsigned char* ip){
	e1 = (EIPTYPE1*)eip;
	if(debug_eip_host && debug_eip_host_in){ // debug block
		printf(" INFO: IP dst vs EIP dst: \n");
		memdmp((char*)(ip+12), 8);printf(" vs ");memdmp((char*)(&e1->dst), 8);
		printf("\n");
	}
	if(debug_eip_host){
		printf("PERF: TYPE1: %u\n", e1->proc_time);
	}
	if(eip_flow_verify(eip)){
		FLOW_CACHE_RECORD *flow;
		ADDR_CACHE_RECORD *addr;
		bug("eip_validate_type1:");flow = flowcache_lookup((FLOW_CACHE_KEY*)&((iphdr*)ip)->saddr); //the key struct is the ip saddr and daddr
		bug("eip_validate_type1:");addr = addrcache_lookup((ADDR_CACHE_KEY*)&((iphdr*)ip)->saddr); //the key struct is the ip saddr and daddr
		if(flow){
			if(debug_eip_host) printf("...expunging old incoming flow\n");
			// remove current flow record
			bug("eip_validate_type1:");flowcache_del(flow);
		}
		bug("eip_validate_type1:");flowcache_add(e1);
		// store inbound t-addresses in the address cache
		bug("eip_validate_type1:");addrcache_add(&e1->src);
		bug("eip_validate_type1:");addrcache_add(&e1->dst);
		return EIP_VALID;
	}
	return EIP_INVALID; // default
}

int eip_validate_type2(const unsigned char* eip, const unsigned char* ip, const int hasContinuation){
    FLOW_CACHE_RECORD *flow;
    flow = flowcache_lookup((FLOW_CACHE_KEY*)&((iphdr*)ip)->saddr); //the key struct is the ip saddr and daddr
    if(flow){
		unsigned int pktBytes = ntohs(((iphdr*)ip)->tot_len);
		flow->bytes += pktBytes; // add total bytes to accounting
		if(hasContinuation){
			e2c = (EIPTYPE2C*)eip;
			bug("eip_validate_type2:");addrcache_add(&e2c->cont_taddr); // cache the continuation address
			if(debug_eip_host && debug_eip_host_in){ printf("...cont_addr=");memdmp((char*)&e2c->cont_taddr,8);printf("\n"); }//debug
			if(debug_eip_host){
				printf("PERF: TYPE2C: %u\n", e2c->proc_time);
			}
		} else {
			if(debug_eip_host){
				e2 = (EIPTYPE2*)eip;
				printf("PERF: TYPE2: %u\n", e2->proc_time);
			}
		}
		return EIP_VALID;
	}
	return EIP_INVALID;
}

int eip_validate_type3(const unsigned char* eip, const unsigned char* ip, const int hasContinuation){
    FLOW_CACHE_RECORD *flow;
    bug("eip_validate_type3:");flow = flowcache_lookup((FLOW_CACHE_KEY*)&((iphdr*)ip)->saddr); //the key struct is the ip saddr and daddr
    if(flow){
		if(eip_flow_verify(eip)){
			unsigned int pktBytes = ntohs(((iphdr*)ip)->tot_len);
			flow->bytes += pktBytes; // add total bytes to accounting
			if(hasContinuation){//
				e3c = (EIPTYPE3C*)eip;
				bug("eip_validate_type3:");addrcache_add(&e3c->cont_taddr);
				if(debug_eip_host && debug_eip_host_in){ printf("...cont_addr=");memdmp((char*)&e3c->cont_taddr,8);printf("\n"); }//debug
				if(debug_eip_host){
					printf("PERF: TYPE3C: %u\n", e3c->proc_time);
				}
			} else {
				if(debug_eip_host){
					e3 = (EIPTYPE3*)eip;
					printf("PERF: TYPE3: %u\n", e3->proc_time);
				}
			}
			return EIP_VALID;
		}
	}
	//printf("...added %d bytes to flow\n", iplen);
	//if(flow->bytes > flow->taddr.vbytes) printf("...t-address overbyte: %d > %d\n", flow->bytes, flow->taddr.vbytes);
	return EIP_INVALID;
}

//
// header preparation functions
//
void eip_send_type1(unsigned char *headers, const unsigned char *pkt, const unsigned int len, const int useContinuation){
	e1 = (EIPTYPE1*)headers;
    e1->type[0] = EIP_TYPE_SETUP; // set type byte
    eip_nonce(e1->flowid, EIP_FLOW_ID_LEN); // get new flow id
    e1->flowttl = time(NULL) + EIP_FLOW_TTL;
    e1->timestamp = time(NULL);
    eip_flow_sign(e1->sig, e1->flowid, (char*)&e1->flowttl, (char*)&e1->timestamp, NULL);
    eip_fetch_src_taddr((iphdr*)pkt, &e1->src);
    bug("eip_send_type1:");ADDR_CACHE_RECORD* addr = addrcache_lookup((ADDR_CACHE_KEY*)(pkt+IPHDR_ADDR_OFFSET));
    if(useContinuation && addr){
		memcpy(&e1->dst, &addr->taddr, sizeof(TADDRESS));
		addr->sent = 1; // set flag to indicate that received continuation address has been used
	} else {
		eip_fetch_dst_taddr((iphdr*)pkt, &e1->dst); // get the dstT from the flow cache
	}
	// check to see if this flow needs to be expunged
	FLOW_CACHE_RECORD *flow;
	FLOW_CACHE_KEY key; // key, which is the reversed saddr/daddr combination, because this is outgoing
	memcpy(&key, e1->dst.src, EIP_NET_ADDR_LEN*2);
	bug("eip_send_type1:");flow = flowcache_lookup(&key); //the key struct is the ip saddr and daddr
	if(flow){
		if(debug_eip_host) printf("...expunging old outgoing flow\n");
		// remove current flow record
		bug("eip_send_type1:");flowcache_del(flow);
	}
    bug("eip_send_type1:");flowcache_add(e1); // add flow to cache
}

void eip_send_type2(unsigned char *headers, const unsigned char *pkt, const unsigned int len, FLOW_CACHE_RECORD* flow, const int addContinuation){
	e2 = (EIPTYPE2*)headers;      // point struct at buffer    
	e2->type[0] = EIP_TYPE_REUSE; // set type byte    
	e2->continuation = 0;  
	eip_flow_accounting(flow, (iphdr*)pkt);  // update flow cache validity constraints
	memcpy(e2->flowid, flow->flowid, EIP_FLOW_ID_LEN); // copy flowid from cached flow
	if(addContinuation){
		int need_to_send = 0;
		bug("eip_send_type2:");ADDR_CACHE_RECORD* addr = addrcache_lookup_reverse((ADDR_CACHE_KEY*)(pkt+IPHDR_ADDR_OFFSET));
		if(addr){
			if(addr->sent == 0){
				need_to_send = 1;
				if(debug_eip_host && debug_eip_host_out) printf("...need to send cont b/c send=0\n");
			}
		} else {
			if(debug_eip_host && debug_eip_host_out) printf("...need to send cont b/c addr DNE\n");
			need_to_send = 1;
		}
		if(need_to_send){
			e2c = (EIPTYPE2C*)headers;
			e2c->continuation = 1;     // set bit indicating this datagram includes a continuation t-address
			eip_generate_taddr((iphdr*)pkt, &e2c->cont_taddr); // generate new address for communication to this host
			if(debug_eip_host && debug_eip_host_out){ printf("...cont_addr=");memdmp((char*)&e2c->cont_taddr,8);printf("\n"); }//debug
			bug("eip_send_type2:");addrcache_add_sent(&e2c->cont_taddr);
		}
	}
}

void eip_send_type3(unsigned char *headers, const unsigned char *pkt, const unsigned int len, FLOW_CACHE_RECORD* flow, const int addContinuation){
	// flow has expired, so renew the flow
	e3 = (EIPTYPE3*)headers;       // point struct at buffer
	e3->type[0] = EIP_TYPE_RENEW;  // set type byte
	e3->continuation = 0;
	memcpy(e3->old_flowid, flow->flowid, EIP_FLOW_ID_LEN); // copy flow id
	eip_nonce(e3->new_flowid, EIP_FLOW_ID_LEN); // generate new flowid
	e3->flowttl = time(NULL) + EIP_FLOW_TTL;   // set new flow ttl
	e3->timestamp = time(NULL);
	eip_flow_sign(e3->sig, e3->new_flowid, (char*)&e3->flowttl, (char*)&e3->timestamp, NULL);  // sign new flow
	flowcache_update(flow, e3); // update flow cache
	if(addContinuation){
		int need_to_send = 0;
		bug("eip_send_type3:");ADDR_CACHE_RECORD* addr = addrcache_lookup_reverse((ADDR_CACHE_KEY*)(pkt+IPHDR_ADDR_OFFSET));
		if(addr){
			if(addr->sent == 0){
				need_to_send = 1;
				if(debug_eip_host && debug_eip_host_out) printf("...need to send cont b/c send=0\n");
			}
		} else {
			if(debug_eip_host && debug_eip_host_out) printf("...need to send cont b/c addr DNE\n");
			need_to_send = 1;
		}
		if(need_to_send){
			e3c = (EIPTYPE3C*)headers;
			e3c->continuation = 1;     // set bit indicating this datagram includes a continuation t-address
			eip_generate_taddr((iphdr*)pkt, &e3c->cont_taddr); // generate new address for communication to this host
			if(debug_eip_host && debug_eip_host_out){ printf("...cont_addr=");memdmp((char*)&e3c->cont_taddr,8);printf("\n"); }//debug
			bug("eip_send_type3:");addrcache_add(&e3c->cont_taddr);
			bug("eip_send_type3:");addr = addrcache_lookup((ADDR_CACHE_KEY*)(&e3c->cont_taddr));
			addr->sent = 1; // set the flag indicating that the continuation has been sent for this flow 
		}
	}
}

int eip_use_continuation(FLOW_CACHE_RECORD *flow){
	int rval = (flow->bytes > flow->taddr.vbytes - EIP_TADDRESS_CONT_BYTES_THRESH);
	//if(debug_eip_host){
		printf("...eip_use_continuation: flow bytes=%u  thresh bytes=%u use_continuation=%d", flow->bytes, flow->taddr.vbytes - EIP_TADDRESS_CONT_BYTES_THRESH, rval);
	//}
	return rval;
}

//
// threshold validity functions
//
int eip_flow_isValid(FLOW_CACHE_RECORD* flow){
	if (time(NULL) + EIP_FLOW_TTL_THRESH < flow->ttl){		
		return 1;
	} else {
		if(debug_eip_host) printf("...flow over TTL threshold (%lu > %lu)\n", time(NULL)+EIP_FLOW_TTL_THRESH, flow->ttl);//debug
	}
	return 0;
}
int eip_capability_isValid(FLOW_CACHE_RECORD* flow){
	if(time(NULL) + EIP_TADDRESS_VTIME_THRESH < flow->taddr.vtime){
		if(flow->bytes + EIP_TADDRESS_VBYTES_THRESH < flow->taddr.vbytes){
			return 1;
		} else {
			if(debug_eip_host) printf("...taddr over byte threshold (%u > %u)\n",flow->bytes + EIP_TADDRESS_VBYTES_THRESH, flow->taddr.vbytes);//debug
			
		}
	} else {
		if(debug_eip_host) printf("...taddr over time threshold (%lu > %lu)\n",time(NULL) + EIP_TADDRESS_VTIME_THRESH,flow->taddr.vtime);//debug
	}
	return 0;
}

int debug_eip_host_crypto = 0;
/**
 * The source t address is a capability to send data to this host.
 * 
 */
void eip_fetch_src_taddr(iphdr *ip, TADDRESS *taddr){
  bug("eip_fetch_src_taddr:");ADDR_CACHE_RECORD* addr = addrcache_lookup((ADDR_CACHE_KEY*)&ip->saddr);
  if(addr){
    memcpy(taddr, &addr->taddr, sizeof(TADDRESS));
    bug("eip_fetch_src_taddr:");addrcache_del(addr);
  } else {
    eip_generate_taddr(ip, taddr); // create srcT and add to datagram
  }
}

/**
 * Fetches the dstT address for type 1 datagram
 * 
 * The destination t-address must be in the cache for communication to
 * occur.  In practice, the dstT must have been placed in temp list by
 * interpreting the DNS response, or by receiving a type-1 datagram from
 * another host which bore a t-address for reverse communication.
 * 
 * In this prototype, delivery of t-addresses over DNS has not yet been 
 * implemented, so if a t-address is not present in the temp list then
 * one will be created.
 * 
 * The destination t-address is a capability to send data to a remote host.
 * 
 */ 
void eip_fetch_dst_taddr(iphdr *ip, TADDRESS *taddr){
  bug("eip_fetch_dst_taddr:");ADDR_CACHE_RECORD* addr = addrcache_lookup((ADDR_CACHE_KEY*)&ip->saddr);
  if(addr){
    memcpy(taddr, &addr->taddr, sizeof(TADDRESS));
    bug("eip_fetch_dst_taddr:");addrcache_del(addr);
  } else {
    //printf("...creating destination t-address\n");//debug
    // This is where we do the stub creation of a t-address, because this condition should never be reached in practice
    //
    // This stub is created using the assumption that all hosts use the
    // same key pair, which is irrelevant to compute cost, but 
    // conveniently allows one host to create a t-address for another.
    // Also, this function uses a temporary IP header which is 
    // constructed to match the IP header that would be present on the
    // destination host, were it constructing the t-address instead.
    //  
    iphdr* tip = (iphdr*)malloc(sizeof(iphdr));
    tip->saddr = ip->daddr;
    tip->daddr = ip->saddr;
    eip_generate_taddr(tip, taddr);
    free(tip);
  }
}

/*
 * Creates srcT address for type 1 datagram
 * 
 * @param ip points to an outbound IP datagram
 * @param e1 points to an EIP type 1 header
 * 
 */ 
void eip_generate_taddr(iphdr *ip, TADDRESS *taddr){
  // this t-address is for communication to this host, so 
  // the destination network address of this t-address is 
  // the source network address of this packet
  memcpy(&taddr->src, &ip->daddr, EIP_NET_ADDR_LEN);
  memcpy(&taddr->dst, &ip->saddr, EIP_NET_ADDR_LEN); 
  taddr->vtime = time(NULL)+EIP_TADDRESS_VTIME; // set expiry time validity constraint
  taddr->vbytes = EIP_TADDRESS_VBYTES;          // set number of bytes validity constraint
  memcpy(&taddr->pubkey, PUBLIC_KEY_PEM, strlen(PUBLIC_KEY_PEM)+1); // copy in the stub public key
  if(!signcache_get(ip, taddr)){ // check the asynchronous signing cache for a ready signature for this key
	eip_taddr_sign(taddr, NULL);                 // else, execute the signature
  }
  //printf("...generated t-addr for ");memdmp((char*)&taddr->src,EIP_NET_ADDR_LEN);printf("->");memdmp((char*)&taddr->dst,EIP_NET_ADDR_LEN);printf("\n");//debug
}

/**
 * Signs flowid, timestamp, and TTL
 * 
 * This function is used for signing flows in type1 and type3 datagrams
 * 
 * @param sig Destination for signature
 * @param flowid Flow ID
 * @param ttl Time to Live
 * @param ts Timestamp
 * @param privkey Private Key for signature
 */ 
void eip_flow_sign(char* sig, char* flowid, char* ttl, char* ts, char* privkey){
  //return; // uncomment to exclude signature operation
  FLOW_SEAL seal;
  memset(&seal, 0, sizeof(FLOW_SEAL));
  if(debug_eip_host_crypto) printf("...eip_flow_sign called\n");//debug
  //const int datalen = EIP_FLOW_ID_LEN + sizeof(uint64_t) + sizeof(uint64_t);//old
  const int datalen = EIP_FLOW_ID_LEN + EIP_TIMESTAMP_LEN * 2;
  char data[datalen]; // allocate buffer for data to be signed
  memcpy(&seal.flowid, flowid, EIP_FLOW_ID_LEN); // copy flowid to signing buffer
  memcpy(&seal.ttl, ttl, EIP_TIMESTAMP_LEN); // copy ttl to buffer
  memcpy(&seal.ts, ts, EIP_TIMESTAMP_LEN);  // copy ts to buffer
  eip_rsa_sign(sig, (char*)&seal, sizeof(FLOW_SEAL));
}

/*
 * Signs t-address
 */ 
void eip_taddr_sign(TADDRESS* taddr, char* privkey){
  //return; // uncomment to exclude signature operation
  if(debug_eip_host_crypto) printf("...eip_taddr_sync sign called\n");//debug
  const int datalen = EIP_NET_ADDR_LEN*2 + sizeof(uint64_t) + sizeof(uint32_t);
  char data[datalen]; // allocate buffer for data to be signed
  memcpy(data, &taddr->src, EIP_NET_ADDR_LEN*2); // copy src and dst addresses (contiguous memory)
  memcpy(data + EIP_NET_ADDR_LEN*2, &taddr->vtime, sizeof(uint64_t));                    // copy vtime to buffer
  memcpy(data + EIP_NET_ADDR_LEN*2 + sizeof(uint64_t), &taddr->vbytes, sizeof(uint32_t)); // copy vbytes to buffer
  eip_rsa_sign(taddr->sig, data, datalen); 
}

/*
 * Updates eip flow record with packet stats with IP packet
 */
int eip_flow_accounting(FLOW_CACHE_RECORD *rec, iphdr* buf){
  rec->bytes += ntohs(buf->tot_len);
} 

/**
 * verify a type 1 or type 3 datagram
 */ 
int eip_flow_verify(const unsigned char* eip_hdr){
  return EIP_VALID; // return immediately without checking validity
                    //
                    // Rabinovich: "the host may do so many things, we don't want to implement any of them"
                    // 
  int rval = EIP_INVALID;
  EIPTYPE1 *e1;
  EIPTYPE3 *e3;
  FLOW_CACHE_RECORD *flow;
  switch(eip_hdr[0]){
    case EIP_TYPE_SETUP:
      e1 = (EIPTYPE1*)eip_hdr;
      if(eip_rsa_verify(e1->flowid, 16+8+8, e1->sig, 128, e1->dst.pubkey, EIP_PUBKEY_LEN)){
        //printf("...flow signature verified\n");
			rval =  EIP_VALID;
      } else {
        //printf("...flow signature invalid\n");
      }
      if(eip_rsa_verify((char*)&e1->dst.src, 4+4+8+4, e1->dst.sig, 128, e1->dst.pubkey, EIP_PUBKEY_LEN)){
        //printf("...dst t-addr signature verified\n");
			rval &=  EIP_VALID;
      } else {
        //printf("...dst t-addr signature invalid\n");
			rval =  EIP_INVALID;        
      }
      break;
    case EIP_TYPE_RENEW:
      //break;
      e3 = (EIPTYPE3*)eip_hdr;
      flow = flowcache_find(e3->old_flowid);
      if(flow){
        if(eip_rsa_verify(e3->new_flowid, 16+8+8, e3->sig, 128, flow->taddr.pubkey, EIP_PUBKEY_LEN)){
          //printf("...new flow signature verified\n");
          rval = EIP_VALID;
        } else {
          //printf("...new flow signature invalid\n");
        }
      } else {
        //printf("...no cached flow was found");
      }
      //return eip_rsa_verify(e3->new_flowid, 128+64+64, e3->sig, 128, e3->dst->pubkey, EIP_PUBKEY_LEN);
      break;
  }
  return EIP_VALID;
}

