/**
 * eip_host.h
 *
 * Headers for EIP host processing
 * 
 */
#ifndef __EIP_HOST_H__
#define __EIP_HOST_H__

#include <stdlib.h> // for malloc
#include <stdint.h> // for uint types
#include <stdio.h>  // for printf
#include <string.h> // for memcpy
#include <arpa/inet.h>   // for htons
#include <netinet/ip.h>  // added for struct iphdr
#include <time.h> // for time
#include "../common/eip-flowcache.h"
#include "../common/eip-addrcache.h"
#include "../common/eip-constants.h"
#include "../common/netfilter-constants.h"
#include "../common/eip-rsa-sign-verify.h"
#include "../common/eip-crypto.h"
#include "./eip-signcache.h"

#define EIP_HEADER_BUFFER_SIZE 16000 // buffer size for constructing new EIP headers
#define EIP_PACKET_BUFFER_SIZE 16000 // buffer size for packet processing

/**
 * eip_validate  Validate the data in an incoming EIP packet
 * @param pkt  Packet data buffer.
 * @param len  Byte length of buffer.
 * @return  Return 1 if the data in pkt is valid, 0 otherwise.
 * 
 */ 
int eip_host_validate(const unsigned char *pkt, const unsigned int len);


/**
 * eip_get_new_headers  Get the number of bytes needed to store the EIP 
 *                      headers for this packet and call appropriate 
 *                      "send" function to places the headers in the
 *                      buffer.
 * @param headers  Buffer for new header data.
 * @param pkt  Original packet data.
 * @param len Byte length of original packet.
 * @return Byte length of header data
 * 
 */ 
unsigned int eip_get_new_headers(unsigned char *headers, const unsigned char *pkt, const unsigned int len);

/**
 * eip_nonce  Places a random bitstring into a buffer
 * @param buf  A buffer.
 * @param len  Number of random bytes to generate.
 * 
 */ 
void eip_nonce(char* buf, const int len);

// validation functions -- return 0 or 1
int eip_validate_type1(const unsigned char* eip, const unsigned char* ip);
int eip_validate_type2(const unsigned char* eip, const unsigned char* ip, const int hasContinuation);
int eip_validate_type3(const unsigned char* eip, const unsigned char* ip, const int hasContinuation);

// header preparation functions -- these place headers in the first buffer
void eip_send_type1(unsigned char *headers, const unsigned char *pkt, const unsigned int len, const int useContinuation);
void eip_send_type2(unsigned char *headers, const unsigned char *pkt, const unsigned int len, FLOW_CACHE_RECORD* flow, const int addContinuation);
void eip_send_type3(unsigned char *headers, const unsigned char *pkt, const unsigned int len, FLOW_CACHE_RECORD* flow, const int addContinuation);

// validity threshold functions -- return 0 or 1
int eip_flow_isValid(FLOW_CACHE_RECORD* flow);
int eip_capability_isValid(FLOW_CACHE_RECORD* flow);

/**
 * eip_use_continuation  Check flow accounting to see if flow has crossed continuation threshold
 * @param flow  The flow to check.
 */
int eip_use_continuation(FLOW_CACHE_RECORD *flow);

//
// recycled functions from earlier versions
//
void eip_fetch_src_taddr(iphdr *ip, TADDRESS *taddr);
void eip_fetch_dst_taddr(iphdr *ip, TADDRESS *taddr);
void eip_generate_taddr(iphdr *ip, TADDRESS *taddr);
void eip_flow_sign(char* sig, char* flowid, char* ttl, char* ts, char* privkey);
void eip_taddr_sign(TADDRESS* taddr, char* privkey);
int eip_flow_accounting(FLOW_CACHE_RECORD *rec, iphdr* buf);
int eip_flow_verify(const unsigned char* eip_hdr);

#endif // __EIP_HOST_H__
