/**
 * EIP host netfilter queue service
 */
 

using namespace std;
#include <iostream>
#include <iomanip>
#include <stdlib.h> // added to get exit() to work
#include <string.h> // added for memcpy
#include <time.h>
#include <netinet/in.h>
#include <netinet/ip.h>  // added for struct iphdr
#include <netinet/tcp.h> // added for struct tcphdr
#include <arpa/inet.h>   // for htons
extern "C" {
  #include <linux/netfilter.h>  /* Defines verdicts (NF_ACCEPT, etc) */
  #include <libnetfilter_queue/libnetfilter_queue.h>
}
// crypto includes
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/rsa.h>
#include <string.h>

// EIP common includes
#include "../common/netfilter-constants.h" // netfilter common values
#include "../common/memdmp.h" // for memdmp debugging function
#include "../common/checksums.c" // checksum functions

extern EVP_MD_CTX *eip_evp_md_ctx; // from eip-rsa-sign-verify

#include "eip_host.h"
#include "eip-signcache.h"

unsigned char eip_header_buffer[EIP_HEADER_BUFFER_SIZE]; // global buffer for new EIP headers
unsigned char new_packet_buffer[NF_PACKET_BUFFER_SIZE]; // global buffer for new packets

// globals for callback function
nfqnl_msg_packet_hdr *nfCallbackPacketHeader = NULL; // Netfilter packet header
uint32_t nfCallbackPacketId = 0; // Netfilter packet ID
int nfCallbackVerdict = 0; // Netfilter Verdict
unsigned int nfCallbackPacketBytes = 0; // byte length of original packet
unsigned char *nfCallbackPacketData = NULL; // pointer to original packet data
int nfCallbackDebug = 0; // debug flag
int nfCallbackDebugDetails = 0; // debug flag
uint64_t inCount = 0;   // inbound packet counter
uint64_t outCount = 0;  // outbound packet counter

// globals for new packet creation / mangling in callback function
unsigned char *nfCallbackNewPacketData = NULL; // pointer to new packet data
unsigned int nfCallbackNewPacketBytes = 0; // byte length of new packet
unsigned int nfCallbackEipBytes = 0; // byte length EIP headers
struct iphdr * nfCallbackIph = NULL; // IP header pointer
struct   tcphdr *nfCallbackTcph = NULL; // TCP header pointer
uint16_t nfCallbackIpTotlen = 0;

/**
 * setMSS(): alters the MSS option of a tcp packet
 * @param pkt  Pointer to the IP header of a packet
 */
void setMSS(unsigned char *pkt){
  struct tcphdr *tcph = (struct tcphdr*)(pkt + 20);
  // now we iterate the tcp options to see if the MSS is set
  unsigned int tcpo_min = 40; // starting address of the TCP options
  unsigned int tcpo_max = tcpo_min + (tcph->doff << 2) - 20 - 4; // starting address of last tcp option; doff<<2 is the whole tcp header len in bytes; -20 removes the standard header; -4 keeps us at least 4 bytes from the end of the options (mss option = 4 bytes)
  uint16_t mss_short = htons(NF_MSS_VALUE); // new MSS value
  for(unsigned int i = tcpo_min; i < tcpo_max; i++){
    if(static_cast<int>(pkt[i]) == 2){ // option type == mss
      if(static_cast<int>(pkt[i+1]) == 4){ // option length == 4
        memcpy(&pkt[i+2], &mss_short, 2);// copy the mss value into the packet buffer
      }
    }
  }
}


/**
 * Callback(): callback function for netfilter hook
 */
 static int Callback(nfq_q_handle *myQueue, struct nfgenmsg *msg, nfq_data *pkt, void *cbData) {
	nfCallbackNewPacketBytes = 0;
	nfCallbackNewPacketData	= NULL;
    if ((nfCallbackPacketHeader = nfq_get_msg_packet_hdr(pkt))) {
        nfCallbackPacketId = ntohl(nfCallbackPacketHeader->packet_id);	 
    }
    nfCallbackPacketBytes = nfq_get_payload(pkt, &nfCallbackPacketData); // get pointer to packet data and # of bytes in packet
    nfCallbackIph = (struct iphdr *)nfCallbackPacketData;
    // print debugging information
    if(nfCallbackDebug && nfCallbackDebugDetails){
        printf("DEBUG: packet length=%d header=", nfCallbackPacketBytes);
        memdmp((char*)nfCallbackPacketData, 20); // print out first 20 bytes of packet
        printf("\n");
    }
    // if this is a TCP packet, we need to add or remove EIP headers
    if(nfCallbackIph->protocol == htons(0x0600)){ // this is a TCP packet
        if(nfq_get_outdev(pkt) > 0){ // this packet is outbound
            outCount ++;
            if(nfCallbackDebug) printf("DEBUG: OUTBOUND #%lu\n", outCount);
            // in EIP host, will insert EIP headers here
            // create a copy of the original packet            
            nfCallbackEipBytes = eip_get_new_headers(eip_header_buffer, nfCallbackPacketData, nfCallbackPacketBytes); // 
            nfCallbackNewPacketBytes = nfCallbackEipBytes + nfCallbackPacketBytes;
            nfCallbackNewPacketData = new_packet_buffer;            
            memset(nfCallbackNewPacketData, 0, nfCallbackNewPacketBytes);
            memcpy(nfCallbackNewPacketData, nfCallbackPacketData, nfCallbackPacketBytes); // copy original content
            memcpy(nfCallbackNewPacketData + nfCallbackPacketBytes, eip_header_buffer, nfCallbackEipBytes); // copy EIP headers
            // update IP total length and store original IP total length
            nfCallbackIph = (struct iphdr*)nfCallbackNewPacketData; // set new ip header pointer to new packet memory location
            nfCallbackTcph = (struct tcphdr*)(nfCallbackNewPacketData + 20); // set new tcp header pointer
            nfCallbackIpTotlen = ntohs(nfCallbackIph->tot_len);
            memcpy(nfCallbackNewPacketData + nfCallbackNewPacketBytes - 2, &(nfCallbackIph->tot_len), 2);
            nfCallbackIph->tot_len = htons(nfCallbackNewPacketBytes);
            if(nfCallbackDebug && nfCallbackDebugDetails) printf("DEBUG: outbound packet expanded from %d to %d\n",nfCallbackPacketBytes,nfCallbackNewPacketBytes);
        }
        if(nfq_get_indev(pkt) > 0){ // this packet is inbound
            inCount ++;
            if(nfCallbackDebug) printf("DEBUG:  INBOUND #%lu\n", inCount);
            // in EIP host, will validate and strip EIP headers here
            if(eip_host_validate(nfCallbackPacketData, nfCallbackPacketBytes)){
				// create a copy of the original packet
				memcpy(&nfCallbackIpTotlen, nfCallbackPacketData + nfCallbackPacketBytes - 2, 2);
				nfCallbackIph->tot_len = nfCallbackIpTotlen;          // restore the original total length
				nfCallbackNewPacketBytes = ntohs(nfCallbackIpTotlen); // new packet is the same as the original packet
				nfCallbackNewPacketData = new_packet_buffer;          // use the new packet buffer
				memset(nfCallbackNewPacketData, 0, nfCallbackPacketBytes);
				memcpy(nfCallbackNewPacketData, nfCallbackPacketData, nfCallbackPacketBytes);
				// restore IP total length
				nfCallbackIph = (struct iphdr*)nfCallbackNewPacketData; // set new ip header pointer to new packet memory location
				nfCallbackTcph = (struct tcphdr*)(nfCallbackNewPacketData + 20); // set new tcp header pointer            
				if(nfCallbackDebug && nfCallbackDebugDetails) printf("DEBUG: inbound packet shrunk from %d to %d\n",nfCallbackPacketBytes,nfCallbackNewPacketBytes);
			} else {
				// ignore -- possible non-EIP packet
				//*
				if(nfCallbackDebug) printf("DEBUG: inbound dropped as invalid\n",nfCallbackPacketBytes,nfCallbackNewPacketBytes);
				nfq_set_verdict(myQueue, nfCallbackPacketId, NF_DROP, 0, NULL); // drop the packet
				return nfCallbackVerdict;
				//*/
			}
        }
        if(nfCallbackNewPacketBytes){
			// recompute IP checksum
			IpChecksum((struct ip*)nfCallbackNewPacketData);        
			// if this is a TCP SYN packet, mangle requested MSS
			if(nfCallbackTcph->syn){
				setMSS(nfCallbackNewPacketData);
			}
			// recompute TCP checksum
			TcpChecksum((struct ip*)nfCallbackIph);
		}
    }
    // return the copy of packet, or NULL pointer, which returns the original packet
    nfCallbackVerdict = nfq_set_verdict(myQueue, nfCallbackPacketId, NF_ACCEPT, nfCallbackNewPacketBytes, nfCallbackNewPacketData);
    return nfCallbackVerdict;
}

 /**
  * main (loop) 
  */
int main(int argc, char **argv) {
    printf("EIP Host started -- press Control-C to stop\n");
    if(argc == 2){
        nfCallbackDebug = 1; // enable debugging based on command-line option
        printf("Debug mode enabled.\n");
    }
    srand(time(NULL)); // seed random number generator
    // crypto init
    OpenSSL_add_all_digests();  // load all known digests
    eip_evp_md_ctx = EVP_MD_CTX_create();  // get a new envelope message digest context
    EVP_MD_CTX_init(eip_evp_md_ctx);       // initialize envelope message digest context
    // signer init
    signcache_setup();

    struct nfq_handle *nfqHandle;
    struct nfq_q_handle *myQueue;
    struct nfnl_handle *netlinkHandle;
    int fd, res;
    char buf[NF_BUFFER_LEN]; // this is the main packet buffer
    // Get a queue connection handle from the module
    if (!(nfqHandle = nfq_open())) {
        cerr << "Error in nfq_open()" << endl;
        exit(-1);
    }
    // Unbind the handler from processing any IP packets
    if (nfq_unbind_pf(nfqHandle, AF_INET) < 0) {
        cerr << "Error in nfq_unbind_pf()" << endl;
        exit(1);
    }
    // Bind this handler to process IP packets...
    if (nfq_bind_pf(nfqHandle, AF_INET) < 0) {
         cerr << "Error in nfq_bind_pf()" << endl;
         exit(1);
    }
    // Install a callback on queue NF_QUEUE_NUM (queue number arbitrary, see notes above)
    if (!(myQueue = nfq_create_queue(nfqHandle,  NF_QUEUE_NUM, &Callback, NULL))) {
        cerr << "Error in nfq_create_queue()" << endl;
        exit(1);
    }
    // Turn on packet copy mode
    if (nfq_set_mode(myQueue, NFQNL_COPY_PACKET, 0xffff) < 0) {
    cerr << "Could not set packet copy mode" << endl;
    exit(1);
  }
  netlinkHandle = nfq_nfnlh(nfqHandle);
  fd = nfnl_fd(netlinkHandle);
  //while ((res = recv(fd, buf, sizeof(buf), 0)) && res >= 0) {
    while(true){
    res = recv(fd, buf, sizeof(buf), 0);
    if(res >= 0) nfq_handle_packet(nfqHandle, buf, res);
    //else cerr << "NFQUEUE_BUFFER_ERROR" << endl;//uncomment this line to see error messages
    // end while receiving traffic
  }
  nfq_destroy_queue(myQueue);
  nfq_close(nfqHandle);
  return 0;
  // end main
}
