/*
 * File: eip_router.cc
 */
#ifndef __EIP_ROUTER_CC__
#define __EIP_ROUTER_CC__

#include "../common/eip-rsa-sign-verify.h"
#include "../common/eip-crypto.h"
#include "../common/eip_timer.h"
#include "eip_router.h"

const int debug_eip_router = 0;     // debug flag

// global pointers for working with EIP datagrams
EIPTYPE1 *e1;
EIPTYPE2 *e2;
EIPTYPE2C *e2c;
EIPTYPE3 *e3;
EIPTYPE3C *e3c;
ROUTER_CACHE_RECORD *rflow;
uint64_t now; // current time in seconds, updated in the callback loop

/**
 * Validate the EIP headers of an inbound packet.
 * @param pkt  Pointer to packet data buffer containing a IP packet with appended EIP headers.
 * @param len  Length of the data in the buffer.
 * @return  EIP_VALID if the packet is valid, EIP_INVALID otherwise.
 * 
 */ 
int eip_router_validate(const unsigned char *pkt, const unsigned int len){
	eip_timer_start(); // processing timer
	uint16_t originalTotalLength;
	memcpy(&originalTotalLength, pkt + len - 2, 2); // get original length from end of packet
	uint16_t originalSize = ntohs(originalTotalLength);
	unsigned char* eip_header = (unsigned char*)(pkt + originalSize);
	if(debug_eip_router) printf("DEBUG: inbound eip type = %d\n", eip_header[0]);//debug
	switch(eip_header[0]){ // first byte of extended packet is EIP type
		case EIP_TYPE_SETUP:
			if(eip_validate_type1(eip_header, pkt)){
				if(debug_eip_router) printf("DEBUG: TYPE 1 valid\n");//debug
				return eip_router_return(EIP_TYPE_SETUP, EIP_CONT_NO, EIP_VALID, eip_header);
			} else {
				if(debug_eip_router) printf("DEBUG: TYPE 1 INVALID\n");//debug
				return EIP_INVALID;
			}
		    break;
		case EIP_TYPE_REUSE:
			if(eip_validate_type2(eip_header, pkt)){
				return eip_router_return(EIP_TYPE_REUSE, EIP_CONT_YES, EIP_VALID, eip_header);
			} else {
				if(debug_eip_router) printf("DEBUG: TYPE 2 INVALID\n");//debug
				return EIP_INVALID;
			}
		    break;
		case EIP_TYPE_RENEW:
			e3 = (EIPTYPE3*) eip_header;
			if(e3->continuation){
				if(debug_eip_router) printf("DEBUG: type 3 has inbound continuation\n");//debug
				if(eip_validate_type3(eip_header, pkt, EIP_CONT_YES)){
					if(debug_eip_router) printf("DEBUG: TYPE 3c valid\n");//debug
					//return EIP_VALID;
					return eip_router_return(EIP_TYPE_RENEW, EIP_CONT_YES, EIP_VALID, eip_header);
				} else {
					if(debug_eip_router) printf("DEBUG: TYPE 3c INVALID\n");//debug
					return EIP_INVALID;
				}
			} else {
				if(eip_validate_type3(eip_header, pkt, EIP_CONT_NO)){
					if(debug_eip_router) printf("DEBUG: TYPE 3 valid\n");//debug
					return EIP_VALID;
				} else {
					if(debug_eip_router) printf("DEBUG: TYPE 3 INVALID\n");//debug
					return EIP_INVALID;
				}
			}
		    break;
	}
	if(debug_eip_router) printf("DEBUG: ***INVALID PACKET***\n");//debug
	return EIP_INVALID; // unreachable
}

int eip_router_return(int datagram_type, int continuation_flag, int return_value, unsigned char* eip){
	switch(datagram_type){
		case EIP_TYPE_SETUP:
			e1 = (EIPTYPE1*)eip;
			eip_timer_stop((uint*)&e1->proc_time);
			break;
		case EIP_TYPE_REUSE:
			e2 = (EIPTYPE2*)eip;
			eip_timer_stop((uint*)&((EIPTYPE2*)eip)->proc_time);
		    /* REMOVED TO TRY SETTING PERF TIME WITHOUT DIFFERENTIATING BETWEEN TYPE2 and TYPE2C
			if(continuation_flag){
				e2c = (EIPTYPE2C*)eip;
				eip_timer_stop((uint*)&e2c->proc_time);
			} else {
				e2 = (EIPTYPE2*)eip;
				eip_timer_stop((uint*)&e2->proc_time);
			}
			*/
			break;
		case EIP_TYPE_RENEW:
			if(continuation_flag){
				e3c = (EIPTYPE3C*)eip;
				eip_timer_stop((uint*)&e3c->proc_time);
			} else {
				e3 = (EIPTYPE3*)eip;
				eip_timer_stop((uint*)&e3->proc_time);
			}
			break;
		default:
			break; // unreachable
	}
	return return_value;
}

//
// validation functions
//
int eip_validate_type1(const unsigned char* eip, const unsigned char* ip){
	e1 = (EIPTYPE1*)eip;
	if(debug_eip_router){ // debug block
		printf(" INFO: IP dst vs EIP dst: \n");
		memdmp((char*)(ip+12), 8);printf(" vs ");memdmp((char*)(&e1->dst), 8);
		printf("\n");
	}
	if(eip_flow_verify(eip)){
		//ADDR_CACHE_RECORD *addr;
		bug("eip_validate_type1:");rflow = routercache_lookup((ROUTER_CACHE_KEY*)&((iphdr*)ip)->saddr); //the key struct is the ip saddr and daddr
		//bug("eip_validate_type1:");addr = addrcache_lookup((ADDR_CACHE_KEY*)&((iphdr*)ip)->saddr); //the key struct is the ip saddr and daddr
		if(rflow){
			if(debug_eip_router) printf("...expunging old incoming flow\n");
			// remove current flow record
			bug("eip_validate_type1:");routercache_del(rflow);
		}
		bug("eip_validate_type1:");routercache_add(e1);
		// store inbound t-addresses in the address cache
		return EIP_VALID;
	}
	return EIP_INVALID; // default
}

int eip_validate_type2(const unsigned char* eip, const unsigned char* ip){
    if(debug_eip_router){		
		bug("eip_validate_type2:");
		memdmp(((EIPTYPE2*)eip)->flowid,EIP_FLOW_ID_LEN);
		printf("\n");
	}
    rflow = routercache_lookup((ROUTER_CACHE_KEY*)((EIPTYPE2*)eip)->flowid); //the key struct flowid
    if(rflow){
		rflow->bytes += ntohs(((iphdr*)ip)->tot_len); // add total bytes to accounting
		if((rflow->bytes < rflow->taddr.vbytes) && (0 < rflow->taddr.vtime)){ // validity check	-- current time not used to avoid clock sync issues
			return EIP_VALID;
		}
		if(debug_eip_router) printf("...type 2 validity check failed");
	}
	return EIP_INVALID;
}

int eip_validate_type3(const unsigned char* eip, const unsigned char* ip, const int hasContinuation){
    e3 = (EIPTYPE3*)eip;
    if(debug_eip_router){		
		bug("eip_validate_type3:");
		memdmp(e3->old_flowid,EIP_FLOW_ID_LEN);
		printf(" > ");
		memdmp(e3->new_flowid,EIP_FLOW_ID_LEN);
		printf("\n");
	}
    rflow = routercache_lookup((ROUTER_CACHE_KEY*)e3->old_flowid); //the key struct is the ip saddr and daddr
    if(rflow){
		printf(" current flow bytes: %u at addr %p\n", rflow->bytes, &(rflow->bytes));//DEBUG
		rflow->bytes += ntohs(((iphdr*)ip)->tot_len); // add total bytes to accounting
		if(eip_flow_verify(eip)){
			return EIP_VALID;
		}
	}
	return EIP_INVALID;
}

/**
 * Verify a type 1 or type 3 datagram.
 * 
 */ 
int eip_flow_verify(const unsigned char* eip_hdr){
  int rval = EIP_INVALID;
  EIPTYPE1 *e1;
  EIPTYPE3 *e3;
  switch(eip_hdr[0]){
    case EIP_TYPE_SETUP:
      e1 = (EIPTYPE1*)eip_hdr;
      if(eip_router_verify(e1->flowid, 16+8+8, e1->sig, 128, e1->dst.pubkey, EIP_PUBKEY_LEN)){
        if(debug_eip_router) printf("...flow signature verified\n");
			rval =  EIP_VALID;
      } else {
        if(debug_eip_router) printf("...flow signature invalid\n");
      }
      if(eip_router_verify((char*)&e1->dst.src, 4+4+8+4, e1->dst.sig, 128, e1->dst.pubkey, EIP_PUBKEY_LEN)){
        if(debug_eip_router) printf("...dst t-addr signature verified\n");
			rval &=  EIP_VALID;
      } else {
        if(debug_eip_router) printf("...dst t-addr signature invalid\n");
			rval =  EIP_INVALID;        
      }
      break;
    case EIP_TYPE_RENEW:
      e3 = (EIPTYPE3*)eip_hdr;
      rflow = routercache_lookup((ROUTER_CACHE_KEY*)e3->old_flowid);
      if(rflow){
        if(eip_router_verify(e3->new_flowid, 16+8+8, e3->sig, 128, rflow->taddr.pubkey, EIP_PUBKEY_LEN)){
          routercache_update(rflow, e3);
          if(debug_eip_router) printf("...new flow signature verified\n");
          rval = EIP_VALID;
        } else {
          if(debug_eip_router) printf("...new flow signature invalid\n");
          routercache_update(rflow, e3);
          rval = EIP_VALID;
        }
      } else {
        if(debug_eip_router) printf("...no cached flow was found\n");
        return EIP_INVALID;
      }
      break;
  }
  return EIP_VALID;
}

int eip_router_verify(char* data, unsigned int datalen, char* sig, unsigned int siglen, char* pubkey, int pubkeylen){
	int i;
	for(i = 0; i < EIP_ROUTER_VERIFICATIONS; i++){
		eip_rsa_verify(data, datalen, sig, siglen, pubkey, pubkeylen);
	}
}

#endif // __EIP_ROUTER_CC__
